"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions 
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Helper functions to aid in parallelizating the creation of sketches using Python's
:code:`multiprocessing`. Also includes :code:`setup_logger` to log statements to both
the console and a log file.
"""
from datetime import datetime
import gc
import logging
from logging.handlers import RotatingFileHandler
from multiprocessing import get_context, Queue
from multiprocessing import shared_memory
from multiprocessing.shared_memory import SharedMemory
from time import sleep
import numpy as np
import psutil
from typing import Callable, Dict, Iterable, List, Tuple, Union

from sketchnu.countmin import CountMin, CountMinLinear, CountMinLog16, CountMinLog8
from sketchnu.heavyhitters import HeavyHitters
from sketchnu.hyperloglog import HyperLogLog


def attach_shared_memory(sketch_type: str, sketch_args: Dict, shm_name: str):
    """
    Create a new sketch and attach it to the shared memory block. Sketch is created by
    passing sketch_args to either CountMin(), HeavyHitters() or HyperLogLog().

    Parameters
    ----------
    sketch_type : str
        Specify the type of sketch to create. Must be "cms" | "hh" | "hll".
    sketch_args : Dict
        The associated arguments to instantiante a new sketch.
    shm_name : str
        Name of a shared memory block that already exist. The new sketch will
        attach to this block.
    
    Returns
    -------
    local_sketch : CountMinLinear | CountMinLog16 | CountMinLog8 | HeavyHitters | HyperLogLog
        A local sketch, relative to this process, that is attached to the shared memory
        block of an existing sketch that has shared_memory = True
    """
    if sketch_type == "cms":
        local_sketch = CountMin(**sketch_args)
    elif sketch_type == "hh":
        local_sketch = HeavyHitters(**sketch_args)
    elif sketch_type == "hll":
        local_sketch = HyperLogLog(**sketch_args)
    else:
        raise TypeError(f"{sketch_type=:} is not handled")

    local_sketch.attach_existing_shm(shm_name)

    return local_sketch


def setup_logger(
    log_console_level: int = logging.INFO,
    log_file_level: int = logging.DEBUG,
    max_bytes: int = 104857600,
    backup_count: int = 5,
    log_file: str = "sketchnu.log",
):
    """
    Create a logger that goes to both the console and a rotating log file.

    Parameters
    ----------
    log_console_level : int, optional
        Minimum log level to send to the console. Default is logging.INFO
    log_file_level : int, optional
        Minimum log level to send to the log file. Default is logging.DEBUG
    max_bytes : int, optional
        Rotate file log once it has gotten to this size. Default is 100MB
    backup_count : int, optional
        Maximum number of old log files to keep. Default is 5
    log_file : str, optional
        Name of the log file. Default is sketchnu.log
    
    Returns
    -------
    logging.Logger
        Logger with the specified levels

    """
    logger = logging.getLogger("sketchnu")
    logger.setLevel(logging.DEBUG)

    # We have already set up this logger before. Don't add more handlers
    # Just set the levels that were requested and return it
    if len(logger.handlers) == 2:
        logger.handlers[0].setLevel(log_console_level)
        logger.handlers[1].setLevel(log_file_level)
        return logger

    formatter = logging.Formatter(
        "%(asctime)s %(name)s %(levelname)s: %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    # Set up the console logging
    console = logging.StreamHandler()
    console.setLevel(log_console_level)
    console.setFormatter(formatter)
    logger.addHandler(console)

    # Set up a file logger
    fh = RotatingFileHandler(log_file, maxBytes=max_bytes, backupCount=backup_count)
    fh.setLevel(log_file_level)
    fh.setFormatter(formatter)
    logger.addHandler(fh)

    return logger


def _fill_queue(
    queue: Queue, items: Iterable, n_workers: int, log_queue: Queue
) -> None:
    """
    Places the items onto the queue and adds poison pill of None for each worker
    """
    for i, item in enumerate(items):
        queue.put(item)
        try:
            item_str = str(item)
        except:
            item_str = str(i + i)
        log_queue.put({"level": "DEBUG", "text": f"{item_str} placed on the queue"})

    for _ in range(n_workers):
        queue.put(None)
    log_queue.put({"level": "INFO", "text": f"All {i+1} items placed on the queue"})

    return None


def _log_worker(
    log_queue: Queue,
    log_console_level: int,
    log_file_level: int,
    max_bytes: int,
    backup_count: int,
    log_file: str,
):
    """
    Worker that will log statements placed onto the log_queue

    Parameters
    ----------
    logger : logging.Logger
    log_queue : multiprocessing.Queue

    Returns
    -------
    None
    """
    logger = setup_logger(
        log_console_level, log_file_level, max_bytes, backup_count, log_file,
    )
    while True:
        q_item = log_queue.get()
        # Process log message
        if q_item is not None:
            try:
                level = q_item.pop("level")
                text = q_item.pop("text")
            except Exception as e:
                logger.error(f"Failed to get level|text from log msg, {q_item}. {e}")
                continue
            if level == "DEBUG":
                logger.debug(text)
            elif level == "INFO":
                logger.info(text)
            elif level == "WARNING":
                logger.warning(text)
            elif level == "CRITICAL":
                logger.critical(text)
            elif level == "ERROR":
                logger.error(text)
            elif level == "FATAL":
                logger.fatal(text)
        else:
            return None


def _worker(
    worker_id: int,
    sketch: Tuple,
    process_q_item: Callable[..., Iterable[bytes]],
    in_queue: Queue,
    log_queue: Queue,
    ngram,
    **kwargs,
):
    """
    Worker takes an item from the queue, uses the function process_q_item()
    to return the list of keys to be added to the cms.

    If sketch is a tuple, then this worker is updating both a cms and hll.

    Parameters
    ----------
    worker_id : int
        Simple integer id for the worker to aid logging
    sketch : Tuple
        How to instantiated sketch(s) using shared memory. Each element is itself
        a tuple with (sketch_type, sketch_args, shm_name) where sketch_type is
        "hll" | "cms", sketch_args are dictionary of arguments to give to HyperLogLog
        or CountMin, and shm_name is the name of the shared memory block to attach the
        instantiated sketch to.
    process_q_item : Callable[..., Iterable[bytes]]
        Function that turns an item from the queue into an Iterator of
        elements (bytes) to be added to the sketch(s).
    in_queue : Queue
        Queue containing items to be processed by the workers
    log_queue : Queue
        Queue to send logging statements to
    ngram : int
        If 0 or None, then calls sketch.add(key). Otherwise calls
        sketch.add_ngram(key, ngram)
    \*\*kwargs :
        Keyword arguments passed to process_q_item(q_item, \*\*kwargs)

    Returns
    -------
    None

    """
    log_queue.put({"level": "INFO", "text": f"WORKER {worker_id:02} is starting"})
    n_records = 0

    # Need to attach shared memory appropriately
    local_sketches = []
    for s in sketch:
        sk = attach_shared_memory(*s)
        local_sketches.append(sk)

    start = datetime.now()
    while True:
        q_item = in_queue.get()
        # Process a queue_item
        if q_item is not None:
            for keys in process_q_item(q_item, **kwargs):
                for local_sketch in local_sketches:
                    if ngram:
                        local_sketch.update_ngram(keys, ngram)
                    else:
                        local_sketch.update(keys)
                n_records += 1

            end = datetime.now()
            speed = n_records / (end - start).total_seconds()
            log_queue.put(
                {
                    "level": "DEBUG",
                    "text": f"WORKER {worker_id:02} has processed "
                    + f"{n_records:,} records at {speed:.3f} records/sec",
                }
            )

        # Poison pill received
        else:
            # Finished processing
            for local_sketch in local_sketches:
                try:
                    # Only the CMS & HH has n_records. Fails if HLL, but we don't care
                    local_sketch.n_added_records[1] += np.uint64(n_records)
                except:
                    pass
                # This should cleanly unattach the shared memory
                del local_sketch

            end = datetime.now()
            speed = n_records / (end - start).total_seconds()
            log_queue.put(
                {
                    "level": "INFO",
                    "text": f"WORKER {worker_id:02} finished "
                    + f"{n_records:,} records at {speed:.3f} records/sec",
                }
            )
            return None


def parallel_add(
    items: Iterable,
    process_q_item: Callable[..., Iterable[Iterable[bytes]]],
    n_workers: int = None,
    ngram: int = None,
    cms_args: Dict = None,
    hh_args: Dict = None,
    hll_args: Dict = None,
    log_file: str = "sketchnu.log",
    log_console_level: int = logging.INFO,
    log_file_level: int = logging.DEBUG,
    max_bytes: int = 104857600,
    backup_count: int = 5,
    **kwargs,
):
    """
    Places `items` onto a queue to be processed by `n_workers` independent spawned
    processes. The function `process_q_item` takes a single item and returns an
    Iterable of records with each record being an Iterable of keys (bytes) to be added
    to the sketch(s). The \*\*kwargs are passed along to `process_q_item` to allow for
    any needed additional parameters. Once all items have been processed, the
    `n_workers` sketch(s) are merged with the final result(s) being returned.

    You must provide at least `hll_args` or `cms_args`. If you provide both, then both
    types of sketches will be processed at the same time while going over the data just
    once.

    Parameters
    ----------
    items : Iterable
        A generator or list of items that will be placed onto a queue and then worked
        by one of the workers in a separate spawned process.
    process_q_item : Callable[..., Iterable[Iterable[bytes]]]
        Function that turns an item from the queue into an Iterable of records with
        each record being an Iterable of keys (bytes) to be added to the sketch(s).
    n_workers : int, optional
        Number of workers to use. Each will update their own sketches which will then
        get merged together to achieve the final sketch(s). If None (default), then set
        to psutil.cpu_count(logical=False)
    ngram : int, optional
        If nonzero, then break each key into ngrams of this size before added to the
        sketch(s)
    cms_args : Dict, optional
        Dictionary containing arguments to instantiate a CountMin.  If None (default)
        then don't create a sketch of this type.
    hh_args : Dict, optional
        Dictionary containing arguments to instantiate a HeavyHitters. If None (default)
        then don't create a sketch of this type.
    hll_args : Dict, optional
        Dictionary containing arguments to instantiate a HyperLogLog. If None (default)
        then don't create a sketch of this type.
    log_file : str, optional
        Filepath to write the logs. Default is sketchnu.log
    log_console_level : int, optional
        Minimum log level to send to the console. Default is INFO
    log_file_level : int, optional
        Minimum log level to send to log_file. Default is DEBUG
    max_bytes: int = 104857600,
        Maximum size of log file before rolling
    backup_count: int = 5,
        Maximum number of log rotating log files to keep
    \*\*kwargs :
        Keyword arguments that get passed to `process_q_item` function
    
    Returns
    -------
    The final sketch(s). If doing more than one sketch, then they are returned as a
    tuple in alphabetical order
    """
    if (cms_args is None) and (hh_args is None) and (hll_args is None):
        raise ValueError("You forgot to provide any sketch arguments")
    if n_workers is None:
        n_workers = max(1, psutil.cpu_count(logical=False))
    if ngram is not None:
        if ngram <= 0:
            raise ValueError(f"ngram is {ngram}. Must be positive")

    ctx = get_context("spawn")
    queue = ctx.Queue(3 * n_workers)
    log_queue = ctx.Queue()

    # Set up logging
    log_process = ctx.Process(
        target=_log_worker,
        args=(
            log_queue,
            log_console_level,
            log_file_level,
            max_bytes,
            backup_count,
            log_file,
        ),
    )
    log_process.start()

    # Start process to fill the queue
    fill_queue_process = ctx.Process(
        target=_fill_queue, args=(queue, items, n_workers, log_queue)
    )
    fill_queue_process.start()

    # Create shared memory sketches
    log_queue.put(
        {"level": "INFO", "text": f"parallel_add: Starting {n_workers} workers"}
    )
    cms_array = []
    hll_array = []
    hh_array = []
    workers = []
    for i in range(n_workers):
        sketch = []
        if cms_args:
            cms_array.append(CountMin(**cms_args, shared_memory=True))
            # Send sketch type, its args, and shared memory block name
            sketch.append(("cms", cms_array[i].args, cms_array[i].shm.name))
        if hll_args:
            hll_array.append(HyperLogLog(**hll_args, shared_memory=True))
            # Send sketch type, its args, and shared memory block name
            sketch.append(("hll", hll_array[i].args, hll_array[i].shm.name))
        if hh_args:
            hh_array.append(HeavyHitters(**hh_args, shared_memory=True))
            # Send sketch type, its args, and shared memory block name
            sketch.append(("hh", hh_array[i].args, hh_array[i].shm.name))
        sketch = tuple(sketch)
        workers.append(
            ctx.Process(
                target=_worker,
                args=(i, sketch, process_q_item, queue, log_queue, ngram),
                kwargs=kwargs,
            )
        )
        workers[i].start()

    # Wait until the queue is finished being populated
    # fill_queue_process.join()

    # Monitor for workers that exit badly
    any_none = True
    while any_none:
        sleep(1)
        any_none = False
        for i, p in enumerate(workers):
            # Still running
            if p.exitcode is None:
                any_none = True
            # Finished but with non-zero exit code, which is bad
            elif p.exitcode != 0:
                msg = f"Worker {i:02} exited badly. {p.exitcode=:}."
                msg = msg + " Possible OOM error"
                for worker in workers:
                    worker.kill()
                if fill_queue_process.exitcode is None:
                    fill_queue_process.kill()
                # Kill the log worker because it is still going
                log_process.kill()
                # Now close all the queues
                queue.close()
                log_queue.close()

    # Wait until the queue is finished being populated
    fill_queue_process.join()

    # Wait for the workers to finish
    for p in workers:
        p.join()

    if cms_args:
        log_queue.put(
            {
                "level": "INFO",
                "text": "parallel_add: Starting successive rounds of merging cms",
            }
        )
        cms_final = parallel_merging(cms_array, log_queue)
    if hh_args:
        log_queue.put(
            {
                "level": "INFO",
                "text": "parallel_add: Starting successive rounds of merging hh",
            }
        )
        hh_final = parallel_merging(hh_array, log_queue)
    if hll_args:
        log_queue.put(
            {
                "level": "INFO",
                "text": "parallel_add: Starting successive rounds of merging hll",
            }
        )
        hll_final = parallel_merging(hll_array, log_queue)

    # Send poison pill to the log_process
    log_queue.put(None)
    # Wait for the log process to finish
    log_process.join()

    if cms_args and hh_args and hll_args:
        return cms_final, hh_final, hll_final
    elif cms_args and hh_args:
        return cms_final, hh_final
    elif cms_args and hll_args:
        return cms_final, hll_final
    elif hh_args and hll_args:
        return hh_final, hll_final
    elif cms_args:
        return cms_final
    elif hh_args:
        return hh_final
    elif hll_args:
        return hll_final


def _merge_worker(sketch1: Tuple[str, Dict, str], sketch2: Tuple[str, Dict, str]):
    """
    Merge sketch2 into sketch1 when sketch1 & sketch2 are in shared memory

    Parameters
    ----------
    sketch1 : Tuple
        (sketch_type, sketch_args, shm_name) that specifies the type of sketch
        "hll" | "cms", the arguments to instantiate a new sketch using HyperLogLog() or
        CountMin(), and the name of the shared memory block that the new sketch will
        attach to.
    sketch2 : Tuple
        (sketch_type, sketch_args, shm_name) that specifies the type of sketch
        "hll" | "cms", the arguments to instantiate a new sketch using HyperLogLog() or
        CountMin(), and the name of the shared memory block that the new sketch will
        attach to.
    
    Returns
    -------
    None
        Side-effect causes sketch2 to merge into sketch1
    """
    # Attach existing shared memory blocks to local sketches
    s1 = attach_shared_memory(*sketch1)
    s2 = attach_shared_memory(*sketch2)

    print(
        f"_merge_worker: Before merge() avail = {psutil.virtual_memory().available/1024**3:.3f}"
    )

    s1.merge(s2)

    print(
        f"_merge_worker: After merge() avail = {psutil.virtual_memory().available/1024**3:.3f}"
    )

    # Clean up
    del s1
    del s2
    gc.collect()

    return None


def parallel_merging(sketch_array: List, log_queue: Queue):
    """
    Merge an array of sketches in successive rounds of pairing. This will use
    at most len(sketch_array) // 2 processes. After merging, the final sketch
    is returned.
    
    Parameters
    ----------
    sketch_array : List
        Array containing sketches to be merged together
    log_queue : Queue
        Log statements to a log Queue
    
    Returns
    -------
    CountMinLinear | CountMinLog16 | CountMinLog8 | HyperLogLog
        The final sketch resulting from merging all of the sketches in the
        sketch_array.

    """
    # CountMinLinear is the parent class so all are instances of it
    if isinstance(sketch_array[0], CountMinLinear):
        sketch_type = "cms"
        sketch_args = sketch_array[0].args
    elif isinstance(sketch_array[0], HeavyHitters):
        sketch_type = "hh"
        sketch_args = sketch_array[0].args
    elif isinstance(sketch_array[0], HyperLogLog):
        sketch_type = "hll"
        sketch_args = sketch_array[0].args
    else:
        raise TypeError("sketch_array[0] is not a HyperLogLog or Count-min sketch")

    for i in range(1, len(sketch_array)):
        if sketch_args != sketch_array[i].args:
            raise TypeError(
                f"{i}th sketch has arguments that are not the same as 0th sketch"
            )

    ctx = get_context("spawn")
    n_to_merge = len(sketch_array)
    while n_to_merge > 1:
        mergers = []
        for i in range(n_to_merge // 2):
            sketch1 = (sketch_type, sketch_args, sketch_array[i * 2].shm.name)
            sketch2 = (sketch_type, sketch_args, sketch_array[i * 2 + 1].shm.name)
            mergers.append(ctx.Process(target=_merge_worker, args=(sketch1, sketch2)))
            mergers[-1].start()

        # Wait for this round of merging to finish
        for p in mergers:
            p.join()
            if p.exitcode < 0:
                raise RuntimeError(f"A _merge_worker had bad {p.exitcode=:}")

        new_sketch_array = []
        for i in range(0, n_to_merge, 2):
            new_sketch_array.append(sketch_array[i])
            # Try to free up memory as we go
            if i + 1 < len(sketch_array):
                sketch_array[i + 1] = None
                gc.collect()

        sketch_array = new_sketch_array
        n_to_merge = len(sketch_array)
        log_queue.put(
            {
                "level": "DEBUG",
                "text": f"parallel_merging: Finished round of merging. {n_to_merge} remaining",
            }
        )

    return sketch_array[0]



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
from collections import Counter
from typing import Iterable
import numpy as np
from pathlib import Path

from sketchnu.countmin import CountMin
from sketchnu.heavyhitters import HeavyHitters
from sketchnu.hyperloglog import HyperLogLog
from sketchnu.helpers import parallel_add
from .zipf import zipf


vocab_size = 1000
vocab = list(
    set([bytes(k) for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)])
)
zipf_p = zipf(vocab_size)
n_data = 25000
data_stream1 = np.random.choice(vocab, n_data, p=zipf_p).tolist()
data_stream2 = np.random.choice(vocab, n_data, p=zipf_p).tolist()


def process_q_item(q_item: Iterable, batch_size: int):
    for i in range(0, len(q_item), batch_size):
        yield q_item[i : i + batch_size]


def test_parallel_add(tmp_path: Path, batch_size: int = 100):
    """
    Test that parallel processing provides the same answers as single threaded. For cms
    we make it wide enough so that the answers are the same for all keys. For hh, the
    order in which things are added can change the values in lhh_count, so we just test
    that we get the same set of keys from parallel & single threaded. We also in sure
    that the n_records() in cms & hh are properly recorded.

    Parameters
    ----------
    tmp_path : Path
        Temporary directory to write the log file
    batch_size : int, optional
        A record is a batch from the data stream of this size. Default is 100
    """
    cms_args = {"cms_type": "linear", "width": 10000, "depth": 8}
    hh_args = {"width": 70, "depth": 4, "max_key_len": 8}
    hll_args = {"p": 16}

    true_count = Counter(data_stream1)
    true_count.update(data_stream2)
    n_records = int(n_data * 2 / batch_size)

    cms, hh, hll = parallel_add(
        [data_stream1, data_stream2],
        process_q_item,
        n_workers=2,
        cms_args=cms_args,
        hh_args=hh_args,
        hll_args=hll_args,
        log_file=tmp_path / "sketchnu.log",
        batch_size=batch_size,
    )

    cms_single = CountMin(**cms_args)
    cms_single.update(data_stream1)
    cms_single.update(data_stream2)
    assert cms.n_records() == n_records
    for i, v in enumerate(vocab):
        assert cms[v] == cms_single[v], f"Failed on {i+1} out of {len(vocab)}, {cms[v]}"

    hh_single = HeavyHitters(**hh_args)
    hh_single.update(data_stream1)
    hh_single.update(data_stream2)
    hh_topk = set([k for k, _ in hh.query(500)])
    hh_single_topk = set([k for k, _ in hh_single.query(500)])
    assert hh.n_records() == n_records
    assert hh_topk == hh_single_topk

    hll_single = HyperLogLog(**hll_args)
    hll_single.update(data_stream1)
    hll_single.update(data_stream2)
    assert hll.query() == hll_single.query()



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions 
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import numpy as np


def zipf(vocab_size, s: int = 1):
    denom = 0.0
    for i in range(vocab_size):
        denom += 1 / (i + 1) ** s

    p = np.zeros(vocab_size, np.float64)
    for k in range(1, vocab_size + 1):
        p[k - 1] = 1 / k ** s / denom

    return p / p.sum()



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Too many problems with the experimental Numba class. Converting this
to a series of Numba functions and a Python classes.

Implementation of count-min sketch using Numba. Three different types have been coded,
but all use conservative updating to help reduce errors. The three types are

* linear : Uses 32-bit linear counters
* log16 : Uses 16-bit log counters
* log8 : Uses 8-bit log counters

Each type is implemented as a Python class, but uses numba functions under the hood
where possible (and faster). The convenience function, :code:`CountMin()` is the
recommended way to instantiate a count-min sketch. The Python classes are
:code:`CountMinLinear`, :code:`CountMinLog16`, and :code:`CountMinLog8`.

Example
-------
To get a linear CountMin with width = 2\*\*20 ::

    from sketchnu.countmin import CountMin

    width = 2**20
    cms = CountMin("linear", width)
    key = "testing".encode("utf-8")
    cms.add(key)
    # To get an estimated count for that key use either
    cms.query(key)
    cms[key]

Each count-min sketch implementation has two special counters that are stored in the
attribute :code:`n_added_records` which is a 1-d array with two elements. The first
records the total number of elements added to the count-min sketch. This is useful when
calculating the error guarantees. Its value can be retrieved with :code:`n_added()`.
The second element is used by :code:`helpers.parallel_add()` to store the number of
records added to the sketch. This value is needed when using a count-min sketch to
calculate the tf-idf value for a given record. The idf piece needs the number of
records which have a given key, i.e. document frequency. This can be provided directly
by the count-min sketch. But you also need to know the total number of records. This
value can be retrieved by :code:`n_records()` if you use :code:`helpers.parallel_add()`
to create the count-min sketch.
"""
import gc
from multiprocessing.shared_memory import SharedMemory
from numba import njit, uint8, uint16, uint32, uint64, float64, types, prange
import numpy as np
from pathlib import Path
from time import sleep
from typing import List, Union

from sketchnu.hashes import fasthash64


@njit(float64(float64, uint64, uint32, uint32))
def _func(base, max_count, num_reserved, uint_max):
    """
    Numba function used to determine the base of the log needed.
    """
    M = float64(max_count) - float64(num_reserved)
    return base ** (uint_max - num_reserved) - M * base + (M - 1.0)


@njit(float64(float64, uint64, uint32, uint32))
def _funcprime(base, max_count, num_reserved, uint_max):
    """
    Numba function used to determine the base of the log needed.
    """
    M = float64(max_count) - float64(num_reserved)
    return uint_max * base ** (uint_max - num_reserved) - M


@njit(float64(uint64, uint32, uint32))
def _find_base(max_count, num_reserved, uint_max):
    """
    Numba function to calcuate the correct base value for log counters using a simple
    Newton method.

    Parameters
    ----------
    max_count: uint64
        Maximum value we want the counter to return
    num_reserved: uint32
        Use linear counting up to this value. Log8 uses 15
    uint_max: uint32
        Maximum value that can be stored. For log8, this is 255
        For log16, this 65536 (2**16 -1)

    Returns
    -------
    float64
        The base to use for the log counters

    Raises
    ------
    ValueError
        If the base is 1.0

    """
    base = float64(np.exp(np.log(max_count) / (uint_max - num_reserved)))

    for i in range(200):
        base = base - _func(base, max_count, num_reserved, uint_max) / _funcprime(
            base, max_count, num_reserved, uint_max
        )
    if base < 1.000000001:
        raise ValueError("Calculated base is 1.0. Raise max_count")
    return base


@njit(float64(uint16, uint16, float64))
def _counter2value(counter, num_reserved, base):
    """
    Numba function to convert a log counter to its corresponding value. Used by both
    CountMinLog16 and CountMinLog8. For log8, the values just get cast to uint16.

    Parameters
    ----------
    counter : np.uint16
        Current value of the counter
    num_reserved : np.uint16
        Number of values reserved for linear counting. After that use log counting
    base : np.float64
        Log base used by the counter

    Returns
    -------
    float64
        Estimated value stored in the log counter
    """
    if counter <= num_reserved:
        return float64(counter)
    else:
        cprime = float64(counter - num_reserved)
        return (base ** cprime - 1.0) / (base - 1.0) + float64(num_reserved)


@njit(types.Tuple((float64, uint64))(float64[:], uint64))
def _rand(rand_batch, rand_ptr):
    """
    Numba function to retrieve a random value from the `rand_batch`. If the pointer
    `rand_ptr` has reached the end of the batch, then create a new batch of 2,048
    random values. When a new batch is created, `rand_batch` is updated as a
    side-effect.

    Parameters
    ----------
    rand_batch : np.ndarray
        1-d array of 2,048 random values with uniform distribution between [0, 1).
    rand_ptr : np.uint64
        Keeps track of where in the rand_batch you are.
    
    Returns
    -------
    rand_value : np.float64
        Random value taken from `rand_batch` found at position = `rand_ptr`
    rand_ptr : np.uint64
        Updated value of the pointer
    """
    if rand_ptr == uint64(2048):
        rand_batch[:] = np.random.rand(2048)
        rand_ptr = uint64(1)
    else:
        rand_ptr += uint64(1)
    return rand_batch[rand_ptr - uint64(1)], rand_ptr


@njit(
    types.Tuple((types.boolean, uint64))(
        uint16, uint16, uint16, float64, float64[:], uint64
    )
)
def _increase_counter(counter, num_reserved, uint_maxval, base, rand_nums, rand_ptr):
    """
    Numba function to randomly determine whether to increase a log counter based upon
    the current counter value. Used by both CountMinLog16 and CountMinLog8. For log8,
    the values just get cast to uint16. If `rand_ptr` is at the end of `rand_nums`,
    then a new batch of `rand_nums` is generated and `rand_ptr` resets to the
    beginning.

    Parameters
    ----------
    counter : np.uint16
        Current value of the counter
    num_reserved : np.uint16
        Number of values reserved for linear counting. After that use log counting
    uint_maxval : np.uint16
        Maximum value that a counter can store. Just the max value of uint of the
        appropriate type (uint8 | uint16)
    base : np.float64
        Log base used by the counter
    rand_nums : np.ndarray, dtype=np.float64
        1-d array of random values
    rand_ptr : np.uint64
        Stores the pointer location into `rand_nums`
    
    Returns
    -------
    bool
        True if counter should be increased. False if it shouldn't
    rand_ptr : uint64
        Updated pointer location into `rand_nums`
    """
    cprime = float64(counter) - float64(num_reserved)
    if counter >= uint_maxval:
        return False, rand_ptr
    elif cprime < 0:
        return True, rand_ptr
    else:
        rand, rand_ptr = _rand(rand_nums, rand_ptr)
        if rand < base ** (-cprime):
            return True, rand_ptr
        else:
            return False, rand_ptr


@njit(
    uint32(
        uint32[:, :],
        uint64[:],
        uint64,
        uint64,
        uint32,
        types.Bytes(types.uint8, 1, "C"),
    )
)
def _query_linear(cms, buckets, width, depth, uint_maxval, key):
    """
    Numba function to query a CountMinLinear for estimated number of times `key` has
    been added to the sketch. Alters the values stored in `buckets` to be the column
    ids that `key` maps to in the sketch.

    Parameters
    ----------
    cms : np.ndarray, dtype=np.uint32
        2-d array of shape (`depth`, `width`) containing the linear counters.
    buckets : np.ndarray, dtype=np.uint64
        1-d array of shape (`depth`,) holding column ids that `key` maps to in cms.
    width : np.uint64
        Number of columns in `cms`
    depth : np.uint64
        Number of rows in `cms`
    uint_maxval : np.uint32
        Maximum uint32 value. That is 2\*\*32 - 1
    key : bytes
    
    Returns
    -------
    int
        Estimated number of times that `key` has been added to the sketch
    """
    min_count = uint_maxval
    for row in range(depth):
        buckets[row] = fasthash64(key, row) % width
        count = cms[row, buckets[row]]
        if count < min_count:
            min_count = count
    return min_count


@njit(
    types.void(
        uint32[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint32,
        types.Bytes(types.uint8, 1, "C"),
    )
)
def _add_linear(cms, n_added_records, buckets, width, depth, uint_maxval, key):
    """
    Numba function to add `key` to the CountMinLinear sketch. Uses conservative
    updating, so it first queries the sketch for the estimated count of `key` which
    causes the buckets to be updated too. Updates `n_added_records[0]` which tracks
    the number of elements added to the sketch.

    Parameters
    ----------
    cms : np.ndarray, dtype=np.uint32
        2-d array of shape (`depth`, `width`) containing the linear counters.
    n_added_records : np.ndarray, dtype=np.uint64
        1-d array of shape (2,) holding special counters for number of elements added
        to the sketch and number of records.
    buckets : np.ndarray, dtype=np.uint64
        1-d array of shape (`depth`,) holding column ids that `key` maps to in cms.
    width : np.uint64
        Number of columns in `cms`
    depth : np.uint64
        Number of rows in `cms`
    uint_maxval : np.uint32
        Maximum uint32 value. That is 2\*\*32 - 1
    key : bytes

    Returns
    -------
    None
    """
    # Track total number of elements added to the sketch
    n_added_records[0] += uint64(1)

    # This gets min_count AND updates buckets
    min_count = _query_linear(cms, buckets, width, depth, uint_maxval, key)

    # Now update only those counters that have min_count
    for row in range(depth):
        count = cms[row, buckets[row]]
        if count == min_count and count < uint_maxval:
            cms[row, buckets[row]] += np.uint32(1)


@njit(
    types.void(
        uint32[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint32,
        types.Bytes(types.uint8, 1, "C"),
        uint64,
    )
)
def _add_ngram_linear(
    cms, n_added_records, buckets, width, depth, uint_maxval, key, ngram
):
    """
    Numba functino to take a given `key`, split it into ngrams of size `ngram`, and add
    the ngrams to a CountMinLinear sketch. If the `key` length is less than `ngram`
    then add the whole `key`.

    Parameters
    ----------
    cms : np.ndarray, dtype=np.uint32
        2-d array of shape (`depth`, `width`) containing the linear counters.
    n_added_records : np.ndarray, dtype=np.uint64
        1-d array of shape (2,) holding special counters for number of elements added
        to the sketch and number of records.
    buckets : np.ndarray, dtype=np.uint64
        1-d array of shape (`depth`,) holding column ids that `key` maps to in cms.
    width : np.uint64
        Number of columns in `cms`
    depth : np.uint64
        Number of rows in `cms`
    uint_maxval : np.uint32
        Maximum uint32 value. That is 2\*\*32 - 1
    key : bytes
        Element to be shingled before adding to the sketch
    ngram : uint64
        ngram size

    Returns
    -------
    None
    """
    key_len = uint64(len(key))
    if key_len <= ngram:
        _add_linear(cms, n_added_records, buckets, width, depth, uint_maxval, key)
    else:
        for i in range(key_len - (ngram - uint64(1))):
            _add_linear(
                cms,
                n_added_records,
                buckets,
                width,
                depth,
                uint_maxval,
                key[i : i + ngram],
            )


@njit(
    types.void(
        uint32[:, :], uint32[:, :], uint64, uint64, uint32, uint64[:], uint64[:],
    ),
    parallel=True,
)
def _merge_linear(
    cms, other_cms, width, depth, uint_maxval, n_added_records, other_n_added_records
):
    """
    Merge the CountMinLinear sketch `other_cms` into the CountMinLinear sketch `cms`.

    Parameters
    ----------
    cms : np.ndarray, dtype=np.uint32
        2-d array of shape (`depth`, `width`) containing the linear counters.
    other_cms : np.ndarray, dtype=np.uint32
        2-d array of shape (`depth`, `width`) containing the linear counters.
    width : np.uint64
        Number of columns in `cms`
    depth : np.uint64
        Number of rows in `cms`
    uint_maxval : np.uint32
        Maximum uint32 value. That is 2\*\*32 - 1
    n_added_records : np.ndarray, dtype=np.uint64
        1-d array of shape (2,) holding special counters for number of elements added
        to the sketch and number of records.
    other_n_added_records : np.ndarray, dtype=np.uint64
        1-d array of shape (2,) holding special counters for number of elements added
        to the sketch and number of records.

    Returns
    -------
    None
    """
    # Merge the count-min sketches
    for row in prange(depth):
        for col in range(width):
            if other_cms[row, col] > uint_maxval - cms[row, col]:
                cms[row, col] = uint_maxval
            else:
                cms[row, col] += other_cms[row, col]
    # Merge the special counters
    n_added_records[0] += other_n_added_records[0]
    n_added_records[1] += other_n_added_records[1]


class CountMinLinear:
    """
    Count-min sketch that uses 32-bit linear counters with conservative updating. A
    given element's maximum count is 2\*\*32 - 1

    Parameters
    ----------
    width : int
        Width of the count-min sketch. Must be non-negative
    depth : int, optional
        Depth of the count-min sketch. Must be non-negative. Default is 8
    shared_memory : bool, optional
        If True, then CountMinLinear is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.

    Attributes
    ----------
    width : np.uint64
        Width of the 2-d array of counters of the count-min sketch
    depth : np.uint64
        Depth of the 2-d array of counters of the count-min sketch
    shared_memory : bool
        Whether `cms` and `n_added_records` are attached to a shared memory block
    cms : np.uint32[:,:]
        2-d array of the counters. Shape = (depth, width)
    n_added_records : np.uint64[:]
        1-d array that holds two special counters. The first is the number of elements
        that have been added to the sketch. Useful for calculating error limits. The
        second is used by helpers.parallel_add() to keep track of the number of records
        that have been processed. Useful if you want to calculate a TF-IDF.
    """

    def __init__(self, width: int, depth: int = 8, shared_memory: bool = False) -> None:
        """
        Initialize a count-min sketch with 32-bit linear counters with conservative
        updating.

        Parameters
        ----------
        width : int
            Width of the count-min sketch. Must be non-negative
        depth : int, optional
            Depth of the count-min sketch. Must be non-negative. Default is 8
        shared_memory : bool, optional
            If True, then CountMinLinear is placed in shared memory. Needed if
            performing multiprocessing as sketchnu.helpers.parallel_add() does.
            Default is False.

        Returns
        -------
        CountMinLinear
        """
        if width <= 0:
            raise ValueError(f"{width=:}. Must be greater than 0")
        if depth <= 0:
            raise ValueError(f"{depth=:}. Must be greater than 0")

        self.width = np.uint64(width)
        self.depth = np.uint64(depth)
        self.uint_maxval = np.uint32(2 ** 32 - 1)

        self.args = {"cms_type": "linear", "width": width, "depth": depth}

        # Stores the column index for a given key. Used frequently, so declaring here
        self.buckets = np.zeros(depth, np.uint64)

        if shared_memory:
            cms_size = int(4 * width * depth)
            n_added_size = 8 * 2
            self.shm = SharedMemory(create=True, size=(cms_size + n_added_size))
            self.cms = np.frombuffer(self.shm.buf[:cms_size], np.uint32).reshape(
                depth, width
            )
            self.n_added_records = np.frombuffer(self.shm.buf[cms_size:], np.uint64)
        else:
            self.cms = np.zeros((depth, width), np.uint32)
            self.n_added_records = np.zeros(2, np.uint64)

    def query(self, key: bytes) -> int:
        """
        Return estimated number of times `key` was added into the count-min sketch

        Parameters
        ----------
        key : bytes
            Element whose estimated count you want returned

        Returns
        -------
        int

        """
        return _query_linear(
            self.cms, self.buckets, self.width, self.depth, self.uint_maxval, key
        )

    def add(self, key: bytes) -> None:
        """
        Add a single key to the count-min sketch and update the counter tracking total
        number of keys added to the count-min sketch. This is in n_added_records[0].

        Parameters
        ----------
        key : bytes
            Element to be added to the sketch

        Returns
        -------
        None
        """
        _add_linear(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            key,
        )

    def update(self, keys: List[bytes]) -> None:
        """
        Add a list of `keys` to the sketch. This follows the convention of
        collections.Counter

        Parameters
        ----------
        keys : List[bytes]
            List of elements to add to the sketch

        Returns
        -------
        None

        """
        for key in keys:
            self.add(key)

    def add_ngram(self, key: bytes, ngram: int) -> None:
        """
        Take a given `key` and shingle it into ngrams of size `ngram` and then
        add the ngrams to the sketch. If the `key` length is less than `ngram`
        then add the whole `key`

        Parameters
        ----------
        key : bytes
            Element to be shingled before adding to the sketch
        ngram : int
            ngram size

        Returns
        -------
        None

        """
        ngram = np.uint64(ngram)
        _add_ngram_linear(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            key,
            ngram,
        )

    def update_ngram(self, keys: List[bytes], ngram: int) -> None:
        """
        Given a list of keys, split each into ngrams of size `ngram`, and then
        add them to the sketch.

        Note
        ----
        Current implementation loops through the keys in Python. Speed testing
        showed that converting the list to numba's typed list was very costly.
        Maybe if that gets faster a pure numba implementation can be done.

        Parameters
        ----------
        keys : List[bytes]
            List of elements to be shingled before adding to the sketch.
        ngram : int
            ngram size

        Returns
        -------
        None

        """
        # Loop through the keys
        for key in keys:
            self.add_ngram(key, ngram)

    def merge(self, other) -> None:
        """
        Merge the count-min sketch `other` into this one.

        Parameters
        ----------
        other : CountMinLinear
            Another CountMinLinear with the same width and depth.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If `other` has different width, depth, or dtype

        """
        if (
            self.width != other.width
            or self.depth != other.depth
            or self.uint_maxval != other.uint_maxval
        ):
            raise TypeError("self and other have different width | depth | type")

        _merge_linear(
            self.cms,
            other.cms,
            self.width,
            self.depth,
            self.uint_maxval,
            self.n_added_records,
            other.n_added_records,
        )

    def save(self, filename: Union[str, Path]) -> None:
        """
        Save the sketch to `filename`

        Parameters
        ----------
        filename: str | Path
            File to save the sketch to disk. This will be a .npz file.

        Returns
        -------
        None

        """
        np.savez(
            filename,
            args=np.array([self.width, self.depth], np.uint64),
            n_added_records=self.n_added_records,
            cms=self.cms,
            dtype=self.cms[0, 0],
        )

    @staticmethod
    def load(filename: Union[str, Path], shared_memory: bool = False):
        """
        Load a saved CountMinLinear stored in `filename`

        Parameters
        ----------
        filename : str | Path
            File path to the saved .npz file
        shared_memory : bool, optional
            If True, load into shared memory. Default is False.

        Returns
        -------
        CountMinLinear
        """
        with np.load(filename) as npzfile:
            args = npzfile["args"]
            cms_dtype = npzfile["dtype"].dtype
            if cms_dtype != np.uint32:
                raise TypeError("Saved sketch is not a CountMinLinear")

            cms = CountMinLinear(*args, shared_memory=shared_memory)
            np.copyto(cms.cms, npzfile["cms"])
            np.copyto(cms.n_added_records, npzfile["n_added_records"])

        return cms

    def attach_existing_shm(self, existing_shm_name: str) -> None:
        """
        Attach this sketch to an existing shared memory block. Useful when working
        within a spawned child process. This creates self.existing_shm which gets
        closed when self.__del__ gets called.

        Parameters
        ----------
        existing_shm_name : str
            Name of an existing shared memory block to attach this sketch to

        Returns
        -------
        None
        """
        existing_shm = SharedMemory(name=existing_shm_name)

        self.cms = np.frombuffer(
            existing_shm.buf[: self.cms.nbytes], self.cms.dtype
        ).reshape(int(self.depth), int(self.width))
        self.n_added_records = np.frombuffer(
            existing_shm.buf[self.cms.nbytes :], np.uint64
        )

        self.existing_shm = existing_shm

    def n_added(self) -> np.uint64:
        """
        This special counter is used to track the total number of elements
        that have been added to the sketch. Useful to check the error guarantees.

        Returns
        -------
        np.uint64
            The number of elements that have been added to the sketch.

        """
        return self.n_added_records[0]

    def n_records(self) -> np.uint64:
        """
        This special counter is used by the sketchnu.helpers.parallel_add() to
        keep track of the number of records that have been added to the sketch.
        This can be used as the numerator of the idf piece of a tf-idf.

        Returns
        -------
        np.uint64
            The number of records that have been added to the sketch.

        """
        return self.n_added_records[1]

    def __getitem__(self, key: bytes) -> int:
        """
        Return estimated number of time `key` was added into the count-min sketch.
        Alias for query(key).
        """
        return self.query(key)

    def __del__(self):
        try:
            if self.shm:
                try:
                    # Need to explicitly del the arrays since they are sharing the
                    # memory block. Without this you get the MemoryError
                    # "cannot close exported pointers exist"
                    del self.cms
                    del self.n_added_records
                    gc.collect()
                    sleep(0.25)
                    self.shm.close()
                    self.shm.unlink()
                except Exception as exc:
                    raise MemoryError(f"Failed to close & unlink: {exc}")
        except AttributeError:
            pass

        try:
            if self.existing_shm:
                try:
                    del self.cms
                    del self.n_added_records
                    gc.collect()
                    sleep(0.25)
                    self.existing_shm.close()
                except Exception as exc:
                    raise MemoryError(f"Failed to close existing_shm: {exc}")
        except AttributeError:
            pass


@njit(
    uint16(
        uint16[:, :],
        uint64[:],
        uint64,
        uint64,
        uint16,
        types.Bytes(types.uint8, 1, "C"),
    )
)
def _query_log16(cms, buckets, width, depth, uint_maxval, key):
    min_count = uint_maxval
    for row in range(depth):
        buckets[row] = fasthash64(key, row) % width
        count = cms[row, buckets[row]]
        if count < min_count:
            min_count = count
    return min_count


@njit(
    uint64(
        uint16[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint16,
        uint16,
        float64,
        float64[:],
        uint64,
        types.Bytes(types.uint8, 1, "C"),
    )
)
def _add_log16(
    cms,
    n_added_records,
    buckets,
    width,
    depth,
    uint_maxval,
    num_reserved,
    base,
    rand_nums,
    rand_ptr,
    key,
):
    # Track total number of elements added to the sketch
    n_added_records[0] += uint64(1)

    # This gets min_count AND updates buckets
    min_count = _query_log16(cms, buckets, width, depth, uint_maxval, key)
    increase, rand_ptr = _increase_counter(
        min_count, num_reserved, uint_maxval, base, rand_nums, rand_ptr
    )
    if increase:
        # Now update only those counter that have min_count
        for row in range(depth):
            count = cms[row, buckets[row]]
            if count == min_count:
                cms[row, buckets[row]] += np.uint16(1)

    return rand_ptr


@njit(
    uint64(
        uint16[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint16,
        uint16,
        float64,
        float64[:],
        uint64,
        types.Bytes(types.uint8, 1, "C"),
        uint64,
    )
)
def _add_ngram_log16(
    cms,
    n_added_records,
    buckets,
    width,
    depth,
    uint_maxval,
    num_reserved,
    base,
    rand_nums,
    rand_ptr,
    key,
    ngram,
):
    key_len = uint64(len(key))
    if key_len <= ngram:
        rand_ptr = _add_log16(
            cms,
            n_added_records,
            buckets,
            width,
            depth,
            uint_maxval,
            num_reserved,
            base,
            rand_nums,
            rand_ptr,
            key,
        )
    else:
        for i in range(key_len - (ngram - uint64(1))):
            rand_ptr = _add_log16(
                cms,
                n_added_records,
                buckets,
                width,
                depth,
                uint_maxval,
                num_reserved,
                base,
                rand_nums,
                rand_ptr,
                key[i : i + ngram],
            )
    return rand_ptr


@njit(
    types.void(
        uint16[:, :],
        uint16[:, :],
        uint64,
        uint64,
        uint64,
        uint16,
        uint16,
        float64,
        uint64[:],
        uint64[:],
    ),
    parallel=True,
)
def _merge_log16(
    cms,
    other_cms,
    width,
    depth,
    max_count,
    uint_maxval,
    num_reserved,
    base,
    n_added_records,
    other_n_added_records,
):
    """
    Merge other_registers into registers

    Parameters
    ----------
    cms : np.ndarray, dtype=uint16, shape=(depth, width)
    other_cms : np.ndarray, dtype=uint16, shape=(depth, width)
    width : uint64
    depth : uint64
    max_count : uint64
    uint_maxval : uint16
    num_reserved : uint16
    base : float64
    n_added_records : np.ndarray, dtype=uint64, shape=(2,)
    other_n_added_records : np.ndarray, dtype=uint64, shape=(2,)

    Returns
    -------
    None
    """
    # Need to take care of various cases with log counters
    for row in prange(depth):
        for col in range(width):
            # Get what the combined value should be
            v = _counter2value(cms[row, col], num_reserved, base) + _counter2value(
                other_cms[row, col], num_reserved, base
            )
            # If less than num_reserved, then c = v
            if v <= num_reserved:
                cms[row, col] = uint16(v)
            elif v >= max_count:
                cms[row, col] = uint_maxval
            else:
                cprime = np.log((v - num_reserved) * (base - 1.0) + 1.0) / np.log(base)
                cprime = uint16(cprime)
                clower = cprime + num_reserved
                vlower = _counter2value(clower, num_reserved, base)
                vhigher = _counter2value(clower + uint16(1), num_reserved, base)
                delta = v - vlower
                if delta / (vhigher - vlower) <= 0.5:
                    cms[row, col] = clower
                else:
                    cms[row, col] = clower + uint16(1)

    # Merge the special counters
    n_added_records[0] += other_n_added_records[0]
    n_added_records[1] += other_n_added_records[1]


class CountMinLog16(CountMinLinear):
    """
    Count-min sketch that uses 16-bit log counters with conservative updating.

    Parameters
    ----------
    width : int
        Width of the count-min sketch. Must be non-negative.
    depth : int, optional
        Depth of the count-min sketch. Must be non-negative. Default is 8.
    max_count : int, optional
        The maximum value we want to count up to for any given key. Default is
        2\*\*32 -1 (4,294,967,295).
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 1023. This must be less than
        65,535 (2\*\*16 - 1).
    shared_memory : bool, optional
        If True, then CountMinLinear is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.

    Attributes
    ----------
    width : np.uint64
        Width of the 2-d array of counters of the count-min sketch
    depth : np.uint64
        Depth of the 2-d array of counters of the count-min sketch
    max_count : int, optional
        The maximum value we want to count up to for any given key. Default is
        2\*\*32 -1 (4,294,967,295).
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 1023. This must be less than
        65,535 (2\*\*16 - 1).
    shared_memory : bool
        Whether `cms` and `n_added_records` are attached to a shared memory block
    cms : np.uint32[:,:]
        2-d array of the counters. Shape = (depth, width)
    n_added_records : np.uint64[:]
        1-d array that holds two special counters. The first is the number of elements
        that have been added to the sketch. Useful for calculating error limits. The
        second is used by helpers.parallel_add() to keep track of the number of records
        that have been processed. Useful if you want to calculate a TF-IDF.
    base : np.float64
        Calculated base for the log counters. Depends affected by max_count and
        num_reserved.
    """

    def __init__(
        self,
        width: int,
        depth: int = 8,
        max_count=4294967295,
        num_reserved=1023,
        shared_memory: bool = False,
    ) -> None:
        """
        Initialize a count-min sketch with 16-bit log counters with conservative
        updating.

        Parameters
        ----------
        width : int
            Width of the count-min sketch. Must be non-negative.
        depth : int, optional
            Depth of the count-min sketch. Must be non-negative. Default is 8.
        max_count : int, optional
            The maximum value we want to count up to for any given key. Default is
            2\*\*32 -1 (4,294,967,295).
        num_reserved : int, optional
            Perform linear counting for values [0, num_reserved]. After that use log
            counters. This gives more precise estimates for the number of times a key
            is seen for counts <= num_reserved. Default is 1023. This must be less than
            65,535 (2\*\*16 - 1).
        shared_memory : bool, optional
            If True, then CountMinLinear is placed in shared memory. Needed if
            performing multiprocessing as sketchnu.helpers.parallel_add() does.
            Default is False.

        Returns
        -------
        CountMinLog16
        """
        if width <= 0:
            raise ValueError(f"{width=:}. Must be greater than 0")
        if depth <= 0:
            raise ValueError(f"{depth=:}. Must be greater than 0")
        if num_reserved >= 65535:
            raise ValueError(f"{num_reserved=:,}. Must be less than 65,535")

        self.width = np.uint64(width)
        self.depth = np.uint64(depth)
        self.uint_maxval = np.uint16(2 ** 16 - 1)
        self.max_count = np.uint64(max_count)
        self.num_reserved = np.uint16(num_reserved)

        self.args = {
            "cms_type": "log16",
            "width": width,
            "depth": depth,
            "max_count": max_count,
            "num_reserved": num_reserved,
        }

        # Determine the base of the log counters
        self.base = _find_base(self.max_count, self.num_reserved, self.uint_maxval)

        # Stores the column index for a given key. Used frequently, so declaring here
        self.buckets = np.zeros(depth, np.uint64)

        # Create random numbers in batches of 2048
        rng = np.random.default_rng()
        self.rng = np.random.default_rng(rng.integers(0, 2 ** 63))
        self.rand_ptr = 0
        self.rand_nums = self.rng.random(2048)

        if shared_memory:
            cms_size = int(2 * width * depth)
            n_added_size = 8 * 2
            self.shm = SharedMemory(create=True, size=(cms_size + n_added_size))
            self.cms = np.frombuffer(self.shm.buf[:cms_size], np.uint16).reshape(
                depth, width
            )
            self.n_added_records = np.frombuffer(self.shm.buf[cms_size:], np.uint64)
        else:
            self.cms = np.zeros((depth, width), np.uint16)
            self.n_added_records = np.zeros(2, np.uint64)

    def query(self, key: bytes) -> float:
        """
        Return estimated number of times `key` was added into the count-min sketch

        Parameters
        ----------
        key : bytes
            Element whose estimated count you want returned

        Returns
        -------
        float

        """
        min_counter = _query_log16(
            self.cms, self.buckets, self.width, self.depth, self.uint_maxval, key
        )

        return _counter2value(min_counter, self.num_reserved, self.base)

    def add(self, key: bytes) -> None:
        """
        Add a single key to the count-min sketch and update the counter tracking total
        number of keys added to the count-min sketch. This is in n_added_records[0].

        Parameters
        ----------
        key : bytes
            Element to be added to the sketch
        
        Returns
        -------
        None
        """
        self.rand_ptr = _add_log16(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.rand_nums,
            self.rand_ptr,
            key,
        )

    def add_ngram(self, key: bytes, ngram: int) -> None:
        """
        Take a given `key` and split it into ngrams of size `ngram` and then
        add the ngrams to the sketch. If the `key` length is less than `ngram`
        then add the whole `key`

        Parameters
        ----------
        key : bytes
            Element to be shingled before adding to the sketch
        ngram : int
            ngram size
        
        Returns
        -------
        None

        """
        self.rand_ptr = _add_ngram_log16(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.rand_nums,
            self.rand_ptr,
            key,
            ngram,
        )

    def merge(self, other) -> None:
        """
        Merge the count-min sketch `other` into this one.

        Parameters
        ----------
        other : CountMinLog16
            Another CountMinLog16 with the same parameters.
        
        Returns
        -------
        None

        Raises
        ------
        TypeError
            If `other` has different parameters

        """
        if (
            self.width != other.width
            or self.depth != other.depth
            or self.uint_maxval != other.uint_maxval
            or self.max_count != other.max_count
            or self.num_reserved != other.num_reserved
        ):
            raise TypeError(
                "self and other have different width|depth|type|max_count|num_reserved"
            )

        _merge_log16(
            self.cms,
            other.cms,
            self.width,
            self.depth,
            self.max_count,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.n_added_records,
            other.n_added_records,
        )

    def save(self, filename: Union[str, Path]) -> None:
        """
        Save the count-min sketch to `filename`
        
        Parameters
        ----------
        filename: str | Path
            File to save the hll to disk. This will be a .npz file.
        
        Returns
        -------
        None

        """
        np.savez(
            filename,
            args=np.array([self.width, self.depth, self.max_count, self.num_reserved]),
            n_added_records=self.n_added_records,
            cms=self.cms,
            dtype=self.cms[0, 0],
        )

    @staticmethod
    def load(filename: Union[str, Path], shared_memory: bool = False):
        """
        Load a saved CountMinLog16 stored in `filename`

        Parameters
        ----------
        filename : str | Path
            File path to the saved .npz file
        shared_memory : bool, optional
            If True, load into shared memory. Default is False.
        
        Returns
        -------
        CountMinLog16
        """
        with np.load(filename) as npzfile:
            args = npzfile["args"]
            cms_dtype = npzfile["dtype"].dtype
            if cms_dtype != np.uint16:
                raise TypeError("Saved sketch is not a CountMinLog16")

            cms = CountMinLog16(*args, shared_memory=shared_memory)
            np.copyto(cms.cms, npzfile["cms"])
            np.copyto(cms.n_added_records, npzfile["n_added_records"])

        return cms


@njit(
    uint8(
        uint8[:, :], uint64[:], uint64, uint64, uint8, types.Bytes(types.uint8, 1, "C"),
    )
)
def _query_log8(cms, buckets, width, depth, uint_maxval, key):
    min_count = uint_maxval
    for row in range(depth):
        buckets[row] = fasthash64(key, row) % width
        count = cms[row, buckets[row]]
        if count < min_count:
            min_count = count
    return min_count


@njit(
    uint64(
        uint8[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint8,
        uint8,
        float64,
        float64[:],
        uint64,
        types.Bytes(types.uint8, 1, "C"),
    )
)
def _add_log8(
    cms,
    n_added_records,
    buckets,
    width,
    depth,
    uint_maxval,
    num_reserved,
    base,
    rand_nums,
    rand_ptr,
    key,
):
    # Track total number of elements added to the sketch
    n_added_records[0] += uint64(1)

    # This gets min_count AND updates buckets
    min_count = _query_log8(cms, buckets, width, depth, uint_maxval, key)
    increase, rand_ptr = _increase_counter(
        min_count, num_reserved, uint_maxval, base, rand_nums, rand_ptr
    )
    if increase:
        # Now update only those counter that have min_count
        for row in range(depth):
            count = cms[row, buckets[row]]
            if count == min_count:
                cms[row, buckets[row]] += np.uint8(1)

    return rand_ptr


@njit(
    uint64(
        uint8[:, :],
        uint64[:],
        uint64[:],
        uint64,
        uint64,
        uint8,
        uint8,
        float64,
        float64[:],
        uint64,
        types.Bytes(types.uint8, 1, "C"),
        uint64,
    )
)
def _add_ngram_log8(
    cms,
    n_added_records,
    buckets,
    width,
    depth,
    uint_maxval,
    num_reserved,
    base,
    rand_nums,
    rand_ptr,
    key,
    ngram,
):
    key_len = uint64(len(key))
    if key_len <= ngram:
        rand_ptr = _add_log8(
            cms,
            n_added_records,
            buckets,
            width,
            depth,
            uint_maxval,
            num_reserved,
            base,
            rand_nums,
            rand_ptr,
            key,
        )
    else:
        for i in range(key_len - (ngram - uint64(1))):
            rand_ptr = _add_log8(
                cms,
                n_added_records,
                buckets,
                width,
                depth,
                uint_maxval,
                num_reserved,
                base,
                rand_nums,
                rand_ptr,
                key[i : i + ngram],
            )
    return rand_ptr


@njit(
    types.void(
        uint8[:, :],
        uint8[:, :],
        uint64,
        uint64,
        uint64,
        uint8,
        uint8,
        float64,
        uint64[:],
        uint64[:],
    ),
    parallel=True,
)
def _merge_log8(
    cms,
    other_cms,
    width,
    depth,
    max_count,
    uint_maxval,
    num_reserved,
    base,
    n_added_records,
    other_n_added_records,
):
    """
    Merge other_registers into registers

    Parameters
    ----------
    cms : np.ndarray, dtype=uint8, shape=(depth, width)
    other_cms : np.ndarray, dtype=uint8, shape=(depth, width)
    width : uint64
    depth : uint64
    max_count : uint64
    uint_maxval : uint8
    num_reserved : uint8
    base : float64
    n_added_records : np.ndarray, dtype=uint64, shape=(2,)
    other_n_added_records : np.ndarray, dtype=uint64, shape=(2,)
    
    Returns
    -------
    None
    """
    # Need to take care of various cases with log counters
    for row in prange(depth):
        for col in range(width):
            # Get what the combined value should be
            v = _counter2value(cms[row, col], num_reserved, base) + _counter2value(
                other_cms[row, col], num_reserved, base
            )
            # If less than num_reserved, then c = v
            if v <= num_reserved:
                cms[row, col] = uint8(v)
            elif v >= max_count:
                cms[row, col] = uint_maxval
            else:
                cprime = np.log((v - num_reserved) * (base - 1.0) + 1.0) / np.log(base)
                cprime = uint8(cprime)
                clower = cprime + num_reserved
                vlower = _counter2value(clower, num_reserved, base)
                vhigher = _counter2value(clower + uint8(1), num_reserved, base)
                delta = v - vlower
                if delta / (vhigher - vlower) <= 0.5:
                    cms[row, col] = clower
                else:
                    cms[row, col] = clower + uint8(1)

    # Merge the special counters
    n_added_records[0] += other_n_added_records[0]
    n_added_records[1] += other_n_added_records[1]


class CountMinLog8(CountMinLog16):
    """
    Count-min sketch that uses 8-bit log counters with conservative updating.

    Parameters
    ----------
    width : int
        Width of the count-min sketch. Must be non-negative.
    depth : int, optional
        Depth of the count-min sketch. Must be non-negative. Default is 8.
    max_count : int, optional
        The maximum value we want to count up to for any given key. Default is
        2\*\*32 -1 (4,294,967,295).
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 15. This must be less than
        255 (2\*\*8 - 1).
    shared_memory : bool, optional
        If True, then CountMinLinear is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.

    Attributes
    ----------
    width : np.uint64
        Width of the 2-d array of counters of the count-min sketch
    depth : np.uint64
        Depth of the 2-d array of counters of the count-min sketch
    max_count : int, optional
        The maximum value we want to count up to for any given key. Default is
        2\*\*32 -1 (4,294,967,295).
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 15. This must be less than
        255 (2\*\*8 - 1).
    shared_memory : bool
        Whether `cms` and `n_added_records` are attached to a shared memory block
    cms : np.uint32[:,:]
        2-d array of the counters. Shape = (depth, width)
    n_added_records : np.uint64[:]
        1-d array that holds two special counters. The first is the number of elements
        that have been added to the sketch. Useful for calculating error limits. The
        second is used by helpers.parallel_add() to keep track of the number of records
        that have been processed. Useful if you want to calculate a TF-IDF.
    base : np.float64
        Calculated base for the log counters. Depends affected by max_count and
        num_reserved.
    """

    def __init__(
        self,
        width: int,
        depth: int = 8,
        max_count=4294967295,
        num_reserved=15,
        shared_memory: bool = False,
    ) -> None:
        """
        Initialize a count-min sketch with 8-bit log counters with conservative
        updating.

        Parameters
        ----------
        width : int
            Width of the count-min sketch. Must be non-negative.
        depth : int, optional
            Depth of the count-min sketch. Must be non-negative. Default is 8.
        max_count : int, optional
            The maximum value we want to count up to for any given key. Default is
            2\*\*32 -1 (4,294,967,295).
        num_reserved : int, optional
            Perform linear counting for values [0, num_reserved]. After that use log
            counters. This gives more precise estimates for the number of times a key
            is seen for counts <= num_reserved. Default is 15. This must be less than
            255 (2\*\*8 - 1).
        shared_memory : bool, optional
            If True, then CountMinLinear is placed in shared memory. Needed if
            performing multiprocessing as sketchnu.helpers.parallel_add() does.
            Default is False.
        
        Returns
        -------
        CountMinLog8
        """
        if width <= 0:
            raise ValueError(f"{width=:}. Must be greater than 0")
        if depth <= 0:
            raise ValueError(f"{depth=:}. Must be greater than 0")
        if num_reserved >= 255:
            raise ValueError(f"{num_reserved=:,}. Must be less than 255")

        self.width = np.uint64(width)
        self.depth = np.uint64(depth)
        self.uint_maxval = np.uint8(2 ** 8 - 1)
        self.max_count = np.uint64(max_count)
        self.num_reserved = np.uint8(num_reserved)

        self.args = {
            "cms_type": "log8",
            "width": width,
            "depth": depth,
            "max_count": max_count,
            "num_reserved": num_reserved,
        }

        # Determine the base of the log counters
        self.base = _find_base(self.max_count, self.num_reserved, self.uint_maxval)

        # Stores the column index for a given key. Used frequently, so declaring here
        self.buckets = np.zeros(depth, np.uint64)

        # Create random numbers in batches of 2048
        rng = np.random.default_rng()
        self.rng = np.random.default_rng(rng.integers(0, 2 ** 63))
        self.rand_ptr = 0
        self.rand_nums = self.rng.random(2048)

        if shared_memory:
            cms_size = int(1 * width * depth)
            n_added_size = 8 * 2
            self.shm = SharedMemory(create=True, size=(cms_size + n_added_size))
            self.cms = np.frombuffer(self.shm.buf[:cms_size], np.uint8).reshape(
                depth, width
            )
            self.n_added_records = np.frombuffer(self.shm.buf[cms_size:], np.uint64)
        else:
            self.cms = np.zeros((depth, width), np.uint8)
            self.n_added_records = np.zeros(2, np.uint64)

    def query(self, key: bytes) -> float:
        """
        Return estimated number of times `key` was added into the count-min sketch

        Parameters
        ----------
        key : bytes
            Element whose estimated count you want returned

        Returns
        -------
        float

        """
        min_counter = _query_log8(
            self.cms, self.buckets, self.width, self.depth, self.uint_maxval, key
        )

        return _counter2value(min_counter, self.num_reserved, self.base)

    def add(self, key: bytes) -> None:
        """
        Add a single key to the count-min sketch and update the counter tracking total
        number of keys added to the count-min sketch. This is in n_added_records[0].

        Parameters
        ----------
        key : bytes
            Element to be added to the sketch
        
        Returns
        -------
        None
        """
        self.rand_ptr = _add_log8(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.rand_nums,
            self.rand_ptr,
            key,
        )

    def add_ngram(self, key: bytes, ngram: int) -> None:
        """
        Take a given `key` and split it into ngrams of size `ngram` and then
        add the ngrams to the sketch. If the `key` length is less than `ngram`
        then add the whole `key`

        Parameters
        ----------
        key : bytes
            Element to be shingled before adding to the sketch
        ngram : int
            ngram size
        
        Returns
        -------
        None

        """
        self.rand_ptr = _add_ngram_log8(
            self.cms,
            self.n_added_records,
            self.buckets,
            self.width,
            self.depth,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.rand_nums,
            self.rand_ptr,
            key,
            ngram,
        )

    def merge(self, other) -> None:
        """
        Merge the count-min sketch `other` into this one.

        Parameters
        ----------
        other : CountMinLog8
            Another CountMinLog8 with the same parameters.
        
        Returns
        -------
        None

        Raises
        ------
        TypeError
            If `other` has different parameters

        """
        if (
            self.width != other.width
            or self.depth != other.depth
            or self.uint_maxval != other.uint_maxval
            or self.max_count != other.max_count
            or self.num_reserved != other.num_reserved
        ):
            raise TypeError(
                "self and other have different width|depth|type|max_count|num_reserved"
            )

        _merge_log8(
            self.cms,
            other.cms,
            self.width,
            self.depth,
            self.max_count,
            self.uint_maxval,
            self.num_reserved,
            self.base,
            self.n_added_records,
            other.n_added_records,
        )

    @staticmethod
    def load(filename: Union[str, Path], shared_memory: bool = False):
        """
        Load a saved CountMinLog8 stored in `filename`

        Parameters
        ----------
        filename : str | Path
            File path to the saved .npz file
        shared_memory : bool, optional
            If True, load into shared memory. Default is False.
        
        Returns
        -------
        CountMinLog8
        """
        with np.load(filename) as npzfile:
            args = npzfile["args"]
            cms_dtype = npzfile["dtype"].dtype
            if cms_dtype != np.uint8:
                raise TypeError("Saved sketch is not a CountMinLog8")

            cms = CountMinLog8(*args, shared_memory=shared_memory)
            np.copyto(cms.cms, npzfile["cms"])
            np.copyto(cms.n_added_records, npzfile["n_added_records"])

        return cms


def CountMin(
    cms_type: str,
    width: int,
    depth: int = 8,
    max_count: int = 4294967295,
    num_reserved: int = None,
    shared_memory: bool = False,
) -> Union[CountMinLinear, CountMinLog16, CountMinLog8]:
    """
    Convenience function to instantiate a count-min sketch of the given type.

    Parameters
    ----------
    cms_type : str
        Must be 'linear' | 'log16' | 'log8'
    width : int
        Width of the count-min sketch. Best if you keep width >= n_unique/2
    depth : int, optional
        Depth of the count-min sketch. Sets the number of different hash
        functions used. Probability of exceeding error limits is determined by
        the depth, exp(-depth). Default is 8 which should be fine for most
        circumstances.
    max_count : int, optional
        Maximum value the count a given element may reach. Not used by 'linear'
        type. Default is 2\*\*32 - 1 (4,294,967,295) to match linear type.
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use
        log counters. This gives more precise estimates for the number of times
        an element is seen for counts <= num_reserved. Not used by the 'linear'
        type. Default is None which uses 15 for log8 and 1023 for log16. This
        must be less than the maximum uint value: 65,535 for log16 and 255
        for log8.
    shared_memory : bool, optional
        If True, then count-min sketch is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.
    
    Returns
    -------
    cms : CountMinLinear | CountMinLog16 | CountMinLog8
        A count-min sketch of the requested type and size
    """
    if cms_type == "linear":
        cms = CountMinLinear(width, depth, shared_memory)
    elif cms_type == "log16":
        if num_reserved is None:
            cms = CountMinLog16(width, depth, max_count, shared_memory=shared_memory)
        else:
            cms = CountMinLog16(width, depth, max_count, num_reserved, shared_memory)
    elif cms_type == "log8":
        if num_reserved is None:
            cms = CountMinLog8(width, depth, max_count, shared_memory=shared_memory)
        else:
            cms = CountMinLog8(width, depth, max_count, num_reserved, shared_memory)
    else:
        raise ValueError(f"{cms_type=:}. Must be linear | log16 | log8")

    return cms


def load(
    filename: Union[str, Path], shared_memory: bool = False
) -> Union[CountMinLinear, CountMinLog16, CountMinLog8]:
    """
    Load a saved count-min sketch stored in `filename`

    Parameters
    ----------
    filename: str | Path
        File path to the saved .npz file
    shared_memory : bool
        If True, load into shared memory
        
    Returns
    -------
    CountMinLinear | CountMinLog16 | CountMinLog8
    """
    with np.load(filename) as npzfile:
        cms_dtype = npzfile["dtype"].dtype

    if cms_dtype == np.uint32:
        return CountMinLinear.load(filename, shared_memory)
    elif cms_dtype == np.uint16:
        return CountMinLog16.load(filename, shared_memory)
    elif cms_dtype == np.uint8:
        return CountMinLog8.load(filename, shared_memory)



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""

from collections import Counter
import numpy as np
from pathlib import Path
import pytest

from sketchnu.countmin import CountMin, load
from .zipf import zipf


vocab_size = 25
vocab = list(
    set([bytes(k) for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)])
)
zipf_p = zipf(vocab_size)
n_data = 50000
data_stream = np.random.choice(vocab, n_data, p=zipf_p).tolist()


def test_empty_linear(width: int = 25, depth: int = 8):
    """
    Test that a linear cms starts empty

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("linear", width, depth)
    for key in vocab:
        assert cms[key] == 0.0


def test_empty_log16(width: int = 25, depth: int = 8):
    """
    Test that a log16 cms starts empty

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 1024
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log16", width, depth)
    for data in data_stream:
        assert cms[data] == 0.0


def test_empty_log8(width: int = 25, depth: int = 8):
    """
    Test that a log8 cms starts empty

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 1024
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log8", width, depth)
    for data in data_stream:
        assert cms[data] == 0.0


def test_n_added_linear(width: int = 25, depth: int = 8):
    """
    Test that number of elements added to the sketch is properly tracked

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("linear", width, depth)
    cms.update(data_stream)
    assert cms.n_added() == len(data_stream)


def test_n_added_log16(width: int = 25, depth: int = 8):
    """
    Test that number of elements added to the sketch is properly tracked

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log16", width, depth)
    cms.update(data_stream)
    assert cms.n_added() == len(data_stream)


def test_n_added_log8(width: int = 25, depth: int = 8):
    """
    Test that number of elements added to the sketch is properly tracked

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log8", width, depth)
    cms.update(data_stream)
    assert cms.n_added() == len(data_stream)


def test_query_linear(width: int = 25, depth: int = 8):
    """
    Test that cms.query() is the same as __getitem__()

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("linear", width, depth)
    cms.update(data_stream[:100])
    assert cms.query(data_stream[0]) == cms[data_stream[0]]


def test_query_log16(width: int = 25, depth: int = 8):
    """
    Test that cms.query() is the same as __getitem__()

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log16", width, depth)
    cms.update(data_stream[:5000])
    assert cms.query(data_stream[0]) == cms[data_stream[0]]


def test_query_log8(width: int = 25, depth: int = 8):
    """
    Test that cms.query() is the same as __getitem__()

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log8", width, depth)
    cms.update(data_stream[:5000])
    assert cms.query(data_stream[0]) == cms[data_stream[0]]


def test_update_linear(width: int = 25, depth: int = 8):
    """
    Test that the count-min sketch error quarantees hold. We first assert that
    all estimates are greater than or equal to the true count.  Then we assert
    that with a probability of at least `1 - exp(-depth)` that:

        estimate <= true + N * exp(1) / width

    where `N` is the total number of elements added to the sketch.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 25 (same as vocab_size)
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("linear", width, depth)
    cms.update(data_stream)
    true_count = Counter(data_stream)

    error = np.zeros(vocab_size)
    for i, data in enumerate(vocab):
        error[i] = cms.query(data) - true_count[data]

    max_error = cms.n_added() * np.exp(1) / width
    n_over_max = error[error > max_error].shape[0]

    assert error.min() >= 0.0, f"Minimum error, {error.min():.3f} is negative"
    assert (n_over_max / n_data) < np.exp(
        -depth
    ), f"Exceeded the limit too often, {n_over_max}"


def test_update_log16(
    width: int = 200,
    depth: int = 8,
    max_count: int = 4294967295,
    num_reserved: int = 1023,
    n_trials: int = 20,
    t_stat: float = 3.291,
):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    A log counter has an unbiased estimator of `n` with a variance of
    `(x-1)n(n + 1)/2`, where `x` is the log base.

    This test is for the updating log counters and not for the count-min sketch
    itself. In order to limit the errors introduced by collisions in the
    count-min sketch, we set the width >> vocab_size.

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is much larger than vocab_size to avoid
        collisions in vocab elements. Default is 200
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. Default is 2^32 - 1
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 1023. This must be less than
        65,535 (2^16 - 1).
    n_trials : int, optional
        Number of different experiments to run. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    # Randomly select n_trials keys. Test above & below the num_reserved
    keys = np.random.choice(vocab, n_trials, replace=False).tolist()
    for c in [500, 25000]:
        cms = CountMin("log16", width, depth, max_count, num_reserved)
        true_count = Counter()
        for _ in range(c):
            cms.update(keys)
            true_count.update(keys)

        error = np.zeros(n_trials)
        for i, key in enumerate(keys):
            error[i] = cms.query(key) - true_count[key]

        if error.std() == 0.0:
            assert error.mean() == 0.0, f"std = 0, but mean !=0, {error.mean():.3f}"
        else:
            t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
            # 99.9% confidence level
            assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_update_log8(
    width: int = 200,
    depth: int = 8,
    max_count: int = 4294967295,
    num_reserved: int = 15,
    n_trials: int = 20,
    t_stat: float = 3.291,
):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    A log counter has an unbiased estimator of `n` with a variance of
    `(x-1)n(n + 1)/2`, where `x` is the log base.

    This test is for the updating log counters and not for the count-min sketch
    itself. In order to limit the errors introduced by collisions in the
    count-min sketch, we set the width >> vocab_size.

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is much larger than vocab_size to avoid
        collisions in vocab elements. Default is 200
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. Default is 2^32 - 1
    num_reserved : int, optional
        Perform linear counting for values [0, num_reserved]. After that use log
        counters. This gives more precise estimates for the number of times a key
        is seen for counts <= num_reserved. Default is 15. This must be less than
        255 (2^8 - 1).
    n_trials : int, optional
        Number of different experiments to run. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    # Randomly select n_trials keys. Test above & below the num_reserved
    keys = np.random.choice(vocab, n_trials, replace=False).tolist()
    for c in [12, 25000]:
        cms = CountMin("log8", width, depth, max_count, num_reserved)
        true_count = Counter()
        for _ in range(c):
            cms.update(keys)
            true_count.update(keys)

        error = np.zeros(n_trials)
        for i, key in enumerate(keys):
            error[i] = cms.query(key) - true_count[key]

        if error.std() == 0.0:
            assert error.mean() == 0.0, f"std = 0, but mean !=0, {error.mean():.3f}"
        else:
            t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
            # 99.9% confidence level
            assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_update_ngram_linear(width: int = 200, depth: int = 8):
    """
    Test that ngraming is done right. Default width is set wide enough to ensure
    that we don't have collisions in the count-min sketch.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    data = b"0123456789abcdef"
    for ngram in range(1, len(data) + 2):
        cms = CountMin("linear", width, depth)
        cms.update_ngram([data], ngram)
        counter = Counter()
        for i in range(len(data) - ngram + 1):
            counter.update([data[i : i + ngram]])
        for k, v in counter.items():
            assert cms[k] == v


def test_update_ngram_log16(width: int = 200, depth: int = 8):
    """
    Test that ngraming is done right. Default width is set wide enough to ensure
    that we don't have collisions in the count-min sketch.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    data = b"0123456789abcdef"
    for ngram in range(1, len(data) + 2):
        cms = CountMin("log16", width, depth)
        cms.update_ngram([data], ngram)
        counter = Counter()
        for i in range(len(data) - ngram + 1):
            counter.update([data[i : i + ngram]])
        for k, v in counter.items():
            assert cms[k] == v


def test_update_ngram_log8(width: int = 200, depth: int = 8):
    """
    Test that ngraming is done right. Default width is set wide enough to ensure
    that we don't have collisions in the count-min sketch.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    data = b"0123456789abcdef"
    for ngram in range(1, len(data) + 2):
        cms = CountMin("log8", width, depth)
        cms.update_ngram([data], ngram)
        counter = Counter()
        for i in range(len(data) - ngram + 1):
            counter.update([data[i : i + ngram]])
        for k, v in counter.items():
            assert cms[k] == v


def test_merge_linear(width: int = 200, depth: int = 8):
    """
    Add same data stream into two sketches. Merge them. The results should match the
    true count since width is set large enough that there are no collisions between
    vocab elements (width >> vocab_size)

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms1 = CountMin("linear", width, depth)
    cms2 = CountMin("linear", width, depth)

    cms1.update(data_stream)
    cms2.update(data_stream)
    true_count = Counter(data_stream)
    true_count.update(data_stream)

    cms1.merge(cms2)
    for key in vocab:
        assert cms1.query(key) == true_count[key]


def test_merge_log16(width: int = 200, depth: int = 8, t_stat: float = 3.291):
    """
    Add same data stream into two sketches. Merge them. Results should be close
    to the true_count. width >> vocab_size to avoid collisions in vocab elements

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    cms1 = CountMin("log16", width, depth)
    cms2 = CountMin("log16", width, depth)

    cms1.update(data_stream)
    cms2.update(data_stream)
    true_count = Counter(data_stream)
    true_count.update(data_stream)

    cms1.merge(cms2)
    error = np.zeros(vocab_size)
    for i, key in enumerate(vocab):
        error[i] = cms1[key] - true_count[key]

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean !=0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(vocab_size)))
        # 99.9% confidence level
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_merge_log8(width: int = 200, depth: int = 8, t_stat: float = 3.291):
    """
    Add same data stream into two sketches. Merge them. Results should be close
    to the true_count. width >> vocab_size to avoid collisions in vocab elements

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 200
    depth : int, optional
        Depth of the sketch. Default is 8
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    cms1 = CountMin("log8", width, depth)
    cms2 = CountMin("log8", width, depth)

    cms1.update(data_stream)
    cms2.update(data_stream)
    true_count = Counter(data_stream)
    true_count.update(data_stream)

    cms1.merge(cms2)
    error = np.zeros(vocab_size)
    for i, key in enumerate(vocab):
        error[i] = cms1[key] - true_count[key]

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean !=0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(vocab_size)))
        # 99.9% confidence level
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_saving_linear(tmp_path: Path, width: int = 25, depth: int = 8):
    """
    Test saving and loading a linear count-min sketch

    Parameters
    ----------
    tmp_path : Path
        This is a pytest fixture which will provide a temporary directory unique
        to the test invocation, created in the base temporary directory
    width : int, optional
        Width of the sketch. Default is 25
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("linear", width, depth)
    cms.update(data_stream)
    filename = tmp_path / "test_linear.npz"

    cms.save(filename)
    cms_load = load(filename)

    for data in vocab:
        assert cms[data] == cms_load[data]


def test_saving_log16(tmp_path: Path, width: int = 25, depth: int = 8):
    """
    Test saving and loading a log16 count-min sketch

    Parameters
    ----------
    tmp_path : Path
        This is a pytest fixture which will provide a temporary directory unique
        to the test invocation, created in the base temporary directory
    width : int, optional
        Width of the sketch. Default is 25
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log16", width, depth)
    cms.update(data_stream)
    filename = tmp_path / "test_log16.npz"

    cms.save(filename)
    cms_load = load(filename)

    for data in vocab:
        assert cms[data] == cms_load[data]


def test_saving_log8(tmp_path: Path, width: int = 25, depth: int = 8):
    """
    Test saving and loading a log8 count-min sketch

    Parameters
    ----------
    tmp_path : Path
        This is a pytest fixture which will provide a temporary directory unique
        to the test invocation, created in the base temporary directory
    width : int, optional
        Width of the sketch. Default is 25
    depth : int, optional
        Depth of the sketch. Default is 8
    """
    cms = CountMin("log8", width, depth)
    cms.update(data_stream)
    filename = tmp_path / "test_log8.npz"

    cms.save(filename)
    cms_load = load(filename)

    for data in vocab:
        assert cms[data] == cms_load[data]


def test_max_count_log16(width: int = 25, depth: int = 8, max_count: int = 100000):
    """
    Test that the counters max out at max_count

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is 25
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. If set too low, then
        ValueError is thrown because the base of log counter is ~ 1.0.
        Default is 100,000
    """
    cms = CountMin("log16", width, depth, max_count)
    for _ in range(int(max_count * 1.25)):
        cms.update(vocab)

    for key in vocab:
        assert cms[key] == pytest.approx(max_count)


def test_max_count_log8(width: int = 25, depth: int = 8, max_count: int = 1000):
    """
    Test that the counters max out at max_count

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is 25
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. If set too low, then
        ValueError is thrown because the base of log counter is ~ 1.0.
        Default is 1000
    """
    cms = CountMin("log8", width, depth, max_count)
    for _ in range(int(max_count * 2.0)):
        cms.update(vocab)

    for key in vocab:
        assert cms[key] == pytest.approx(max_count)


def test_num_reserved_log16(
    width: int = 200,
    depth: int = 8,
    max_count: int = 4294967295,
    num_reserved: int = 50,
):
    """
    Test that linear counting is done for counts between [0, num_reserved + 1]

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is 25
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. Default is 2^32 - 1
    num_reserved : int, optional
        Perform linear counting from [0, num_reserved+1]. Default is 50
    """
    cms = CountMin("log16", width, depth, max_count, num_reserved)
    for _ in range(num_reserved + 1):
        cms.update(vocab)
    for key in vocab:
        assert cms[key] == pytest.approx(num_reserved + 1)

    cms.update(vocab)
    # Assert that we are now approximating
    for key in vocab:
        assert cms[key] != pytest.approx(num_reserved + 2)


def test_num_reserved_log8(
    width: int = 200,
    depth: int = 8,
    max_count: int = 4294967295,
    num_reserved: int = 50,
):
    """
    Test that linear counting is done for counts between [0, num_reserved + 1]

    Parameters
    ----------
    width : int, optional
        Width of the count-min sketch. Default is 25
    depth : int, optional
        Depth of the count-min sketch. Default is 8
    max_count : int, optional
        Maximum value that each log counter can store. Default is 2^32 - 1
    num_reserved : int, optional
        Perform linear counting from [0, num_reserved+1]. Default is 50
    """
    cms = CountMin("log8", width, depth, max_count, num_reserved)
    for _ in range(num_reserved + 1):
        cms.update(vocab)
    for key in vocab:
        assert cms[key] == pytest.approx(num_reserved + 1)

    cms.update(vocab)
    # Assert that we are now approximating
    for key in vocab:
        assert cms[key] != pytest.approx(num_reserved + 2)


if __name__ == "__main__":
    test_empty_linear()
    test_empty_log16()
    test_empty_log8()
    test_n_added_linear()
    test_n_added_log16()
    test_n_added_log8()
    test_query_linear()
    test_query_log16()
    test_query_log8()
    test_update_linear()
    test_update_log16()
    test_update_log8()
    test_update_ngram_linear()
    test_update_ngram_log16()
    test_update_ngram_log8()
    test_merge_linear()
    test_merge_log16()
    test_merge_log8()
    test_max_count_log16()
    test_max_count_log8()
    test_num_reserved_log16()
    test_num_reserved_log8()
    # test_saving_linear() # Need to give directory & delete afterwards
    # test_saving_log16()
    # test_saving_log8()



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions 
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Too many problems with the experimental Numba class. Converting this
to a series of Numba functions and a Python class.

Example
-------
To get a HyperLogLog with precision of 16 and seed 0::

    from sketchnu.hyperloglog import HyperLogLog

    hll = HyperLogLog()
    key = 'testing'.encode('utf-8')
    hll.add(key)
    # To get a estimated cardinality
    hll.query()

"""
from multiprocessing.shared_memory import SharedMemory
from numba import njit, uint8, uint64, float64, types
import numpy as np
from pathlib import Path
from typing import List, Union

from sketchnu.hashes import fasthash64
from sketchnu.hll_constants import sub_algorithm_threshold, raw_estimate, bias_data


@njit(float64(uint64, uint64))
def _linear_counting(m, n_zero):
    """
    If there are registers that are zero, then we do linear_counting to
    estimate the cardinality
    
    Parameters
    ----------
    m : int
        Number of registers
    n_zero : int
        Number of registers that equal 0

    Returns
    -------
    float64
        Estimated cardinality when some registers are still zero

    """
    return float64(m) * np.log(float64(m) / float64(n_zero))


@njit(float64(uint8[:], uint64, float64))
def _estimation_function(registers, m, alpha):
    """
    Estimated cardinality with no bias correction

    Returns
    -------
    float64

    """
    total = float64(0.0)
    for r in registers:
        total += 2.0 ** (-float64(r))
    return alpha * float64(m ** 2) / total


@njit(uint8(uint64))
def _n_leading_zeros64(x):
    """
    Calculate the number of leading zeros for a given uint64 integer.
    If numba ever gets a uint6 type, then we could replace uint8.

    Parameters
    ----------
    x : uint64

    Returns
    -------
    uint8
        Number of leading zeros in the bit representation of `x`

    """
    zero = uint64(0)
    n = uint8(64)
    y = x >> uint64(32)
    if y != zero:
        n = n - uint8(32)
        x = y
    y = x >> uint64(16)
    if y != zero:
        n = n - uint8(16)
        x = y
    y = x >> uint64(8)
    if y != zero:
        n = n - uint8(8)
        x = y
    y = x >> uint64(4)
    if y != zero:
        n = n - uint8(4)
        x = y
    y = x >> uint64(2)
    if y != zero:
        n = n - uint8(2)
        x = y
    y = x >> uint64(1)
    if y != zero:
        return n - uint8(2)

    return n - uint8(x)


@njit(float64(uint8[:], uint64, uint64, float64, float64[:], float64[:]))
def _query(registers, m, threshold, alpha, raw_estimate, bias_data):
    """
    Return the estimated cardinality

    Returns
    -------
    float
        Estimated cardinality of elements added to the HyperLogLog so far

    """
    # Get the number of registers that are still zero
    n_zero = m - uint64(np.count_nonzero(registers))

    if n_zero > 0:
        # Start by trying linear counting
        cardinality = _linear_counting(m, n_zero)

        # If above sub-algorithm threshold, switch to estimation function
        # with bias correction
        if cardinality > threshold:
            est = _estimation_function(registers, m, alpha)
            bias = np.interp(est, raw_estimate, bias_data)
            cardinality = est - bias
    else:
        cardinality = _estimation_function(registers, m, alpha)
        # If cardinality is relatively low, then remove estimated bias
        if cardinality <= float64(5 * m):
            bias = np.interp(cardinality, raw_estimate, bias_data)
            cardinality = cardinality - bias

    return cardinality


@njit(types.void(uint8[:], uint64, uint64, uint64, types.Bytes(types.uint8, 1, "C")))
def _add(registers, seed, p, m, key):
    """
    Add a single `key` to the HyperLogLog.
    
    Parameters
    ----------
    registers : uint8[:]
        HyperLogLog's registers
    seed : np.uint64
        Seed for hashing
    p : np.uint64
        HyperLogLog's precision
    m : np.uint64
        Size of HyperLogLog's registers
    key : bytes
        Element to add to the HyperLogLog

    Returns
    -------
    None
        registers has a side-effect of being updated
    """
    hash_val = fasthash64(key, seed)
    # Get the index of the register using the first p bits of the hash
    reg_idx = hash_val & uint64(m - 1)
    # Get the remaining bits for getting number of leading zeros
    bits = hash_val >> p

    # Get number of leading zeros of the 64-bit bits
    # subtract p since the bits is padded with zeros since it is
    # just a 64-p bit value.
    rank = _n_leading_zeros64(bits) - p + 1
    registers[reg_idx] = max(registers[reg_idx], rank)

    return None


@njit(
    types.void(
        uint8[:], uint64, uint64, uint64, types.Bytes(types.uint8, 1, "C"), uint64
    )
)
def _add_ngram(registers, seed, p, m, key, ngram):
    """
    Take a given `key` and split it into ngrams of size `ngram` and then
    add the ngrams to the sketch. If the `key` length is less than `ngram`
    then add the whole `key`

    Parameters
    ----------
    registers : uint8[:]
    seed : uint64
    p : uint64
    m : uint64
    key : bytes
        Element to be shingled before adding to the sketch
    ngram : uint64
        ngram size
    
    Returns
    -------
    None

    """
    key_len = uint64(len(key))
    if key_len <= ngram:
        _add(registers, seed, p, m, key)
    else:
        for i in range(key_len - (ngram - uint64(1))):
            _add(registers, seed, p, m, key[i : i + ngram])


@njit(types.void(uint8[:], uint8[:], uint64))
def _merge(registers, other_registers, m):
    """
    Merge other_registers into registers

    Parameters
    ----------
    registers : np.ndarray, dtype=uint8
        Registers for a HyperLogLog
    other_registers : np.ndarray, dtype=uint8
        Registers for another HyperLogLog that is merged into the first
    m : uint64
        Number of elements in both registers and other_registers
    
    Returns
    -------
    None
        Side-effect that updates `registers`
    """
    for i in range(m):
        registers[i] = max(registers[i], other_registers[i])


class HyperLogLog:
    """
    Implementation of the HyperLogLog++ algorithm which uses a 64-bit hash function and
    corrects bias when cardinality is low.

    Parameters
    ----------
    p : int
        Precision specifies the number of registers (2\*\*p). The larger the p,
        the more accurate the estimated cardinality. Must be between [7, 16]
    seed : int, optional
        Seed passed to the fasthash64 function. Default is 0
    shared_memory : bool, optional
        If True, then HyperLogLog is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.
    
    Attributes
    ----------
    p : int
        Precision of the HyperLogLog. Must be between [7, 16]
    seed : int
        Seed passed to hash function
    m : int
        Number of registers. Equal to 2\*\*p
    shared_memory : bool, optional
        If True, then HyperLogLog is placed in shared memory. Needed if
        performing multiprocessing as sketchnu.helpers.parallel_add() does.
        Default is False.

    """
    def __init__(self, p: int = 16, seed: int = 0, shared_memory: bool = False):
        """
        Instantiate a HyperLogLog. This is an implementation of the HyperLogLog++
        algorithm which uses a 64-bit hash function and corrects bias when cardinality
        is low.

        Parameters
        ----------
        p : int
            Precision specifies the number of registers (2**p). The larger the p,
            the more accurate the estimated cardinality. Must be between [7, 16]
        seed : int, optional
            Seed passed to the fasthash64 function. Default is 0
        shared_memory : bool, optional
            If True, then HyperLogLog is placed in shared memory. Needed if
            performing multiprocessing as sketchnu.helpers.parallel_add() does.
            Default is False.

        Returns
        -------
        HyperLogLog
            The Python class for the HyperLogLog with specified p and seed.

        """
        self.p = np.uint64(p)
        self.seed = np.uint64(seed)

        if self.p > np.uint64(16) or self.p < np.uint64(7):
            raise ValueError("p must be within [7, 16]")

        self.args = {"p": p, "seed": seed}

        # Total number of registers, 2**p
        self.m = np.uint64(1) << self.p
        # Registers will hold the maximum number of (leading zeros + 1) seen
        # for each of the m registers
        if shared_memory:
            self.shm = SharedMemory(create=True, size=int(self.m))
            self.registers = np.frombuffer(self.shm.buf, np.uint8)
        else:
            self.registers = np.zeros(self.m, np.uint8)

        self.alpha = np.float64(0.7213 / (1.0 + 1.079 / self.m))

        # Using [p-7] since we start with p = 7.
        # Gave IndexError: only integers are valid indices
        # Didn't like having np.uint64 and int mixed together
        self.threshold = sub_algorithm_threshold[int(self.p) - 7]
        self.bias_data = bias_data[int(self.p) - 7, :]
        self.raw_estimate = raw_estimate[int(self.p) - 7, :]

    def add(self, key: bytes) -> None:
        """
        Add a single key to the HyperLogLog.
        
        Note
        ----
        Currently, the numba.typed.List needed for the update() are quite slow.
        Speedtest show that you are better off looping through the keys in
        python and add(key) instead of update(numba.typed.List(keys)). Taking
        just over twice as long.

        Parameters
        ----------
        key : bytes
            Element to add to the HyperLogLog

        Returns
        -------
        None
        """
        _add(self.registers, self.seed, self.p, self.m, key)

    def update(self, keys: List[bytes]) -> None:
        """
        Given a list of keys, update the HyperLogLog. This follows the
        convention of collections.Counter in that keys should be a list of
        keys.

        Note
        ----
        Currently, the numba.typed.List needed for the update() are quite slow.
        Speedtest show that you are better off looping through the keys in
        python and add(key) instead of update(numba.typed.List(keys)). Taking
        just over twice as long. If repeated adding the same keys, then update
        is faster.

        Parameters
        ----------
        keys : List[bytes]
            List of elements to add to the HyperLogLog at once

        Returns
        -------
        None

        """
        for key in keys:
            self.add(key)

    def add_ngram(self, key: bytes, ngram: int) -> None:
        """
        Take a given `key` and split it into ngrams of size `ngram` and then
        add the ngrams to the sketch. If the `key` length is less than `ngram`
        then add the whole `key`

        Parameters
        ----------
        key : bytes
            Element to be shingled before adding to the sketch
        ngram : int
            ngram size
        
        Returns
        -------
        None

        """
        ngram = np.uint64(ngram)
        _add_ngram(self.registers, self.seed, self.p, self.m, key, ngram)

    def update_ngram(self, keys: bytes, ngram: int) -> None:
        """
        Given a list of keys, split each into ngrams of size `ngram`, and then
        add them to the sketch.

        Note
        ----
        Speed tests show that you are slightly better off looping through the
        keys in Python and calling add_ngram() instead of update_ngram(). Used
        50k keys, each of 200 bytes, and ngram=4 to find that update_ngram()
        was ~ 6% slower. This is significantly better than add() vs. update().
        If repeated adding the same keys (for testing), then update_ngram() is
        just a little bit faster than add_ngram().

        Parameters
        ----------
        keys : List[bytes]
            List of elements to be shingled before adding to the sketch.
        ngram : int
            ngram size

        Returns
        -------
        None

        """
        # Loop through the keys
        for key in keys:
            self.add_ngram(key, ngram)

    def query(self) -> float:
        """
        Return the estimated cardinality

        Returns
        -------
        float
            Estimated cardinality of elements added to the HyperLogLog so far
        """
        return _query(
            self.registers,
            self.m,
            self.threshold,
            self.alpha,
            self.raw_estimate,
            self.bias_data,
        )

    def merge(self, other) -> None:
        """
        Merge the HyperLogLog `other` into this one.

        Parameters
        ----------
        other : HyperLogLog
            Another HyperLogLog with the same precision and seed
        
        Returns
        -------
        None

        Raises
        ------
        TypeError
            If `other` has different precision or seed from this one

        """
        if self.p != other.p or self.seed != other.seed:
            raise TypeError(f"Precision & seed in other must match self")

        _merge(self.registers, other.registers, self.m)

    def attach_existing_shm(self, existing_shm_name: str) -> SharedMemory:
        """
        Attach this sketch to an existing shared memory block. Useful when working
        within a spawned child process. This creates self.existing_shm which gets
        closed when self.__del__ gets called.

        Parameters
        ----------
        existing_shm_name : str
            Name an existing shared memory block to attach this sketch to
        
        Returns
        -------
        None
        """
        existing_shm = SharedMemory(name=existing_shm_name)
        self.registers = np.frombuffer(existing_shm.buf, np.uint8)

        self.existing_shm = existing_shm

    def save(self, filename: Union[str, Path]) -> None:
        """
        Save the HyperLogLog sketch, hll, to the file, filename
        
        Parameters
        ----------
        filename: str | Path
            File to save the hll to disk. This will be a .npz file.
        
        Returns
        -------
        None

        """
        np.savez(
            filename, args=np.array([self.p, self.seed], np.uint64), hll=self.registers
        )

    @staticmethod
    def load(filename: Union[str, Path], shared_memory: bool = False):
        """
        Load a saved HyperLogLog sketch stored in filename

        Parameters
        ----------
        filename: str | Path
            File path to the saved .npz file
        shared_memory : bool
            If True, copy registers into shared memory
            
        Returns
        -------
        HyperLogLog
        """
        with np.load(filename) as npzfile:
            args = npzfile["args"]
            hll = HyperLogLog(*args, shared_memory=shared_memory)
            np.copyto(hll.registers, npzfile["hll"])

        return hll

    def __del__(self):
        try:
            if self.shm:
                try:
                    self.shm.close()
                    self.shm.unlink()
                except Exception as exc:
                    raise MemoryError(f"Failed to close & unlink: {exc}")
        except AttributeError:
            pass

        try:
            if self.existing_shm:
                try:
                    self.existing_shm.close()
                except Exception as exc:
                    raise MemoryError(f"Failed to close existing_shm: {exc}")
        except AttributeError:
            pass



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions 
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.


Code to replicate determining the bias corrections to low cardinality in the
hyperloglog algorithm.  The paper references doing 200 data points with
5000 experiments at each, for each of the different precision levels which for
sketchnu is [7,16].

We used 200 data points with 50,000 experiments for each precision level. Took
just under 2hrs on a small machine with 4-cores.

Bias is removed if cardinality is between [low, high] where high is 5*2^p

+------+--------+---------+
|   p  |   low  |   high  |
+------+--------+---------+
|   7  |     80 |     640 |
+------+--------+---------+
|   8  |    220 |   1,280 |
+------+--------+---------+
|   9  |    400 |   2,560 |
+------+--------+---------+
|  10  |    900 |   5,120 |
+------+--------+---------+
|  11  |  1,800 |  10,240 |
+------+--------+---------+
|  12  |  3,100 |  20,480 |
+------+--------+---------+
|  13  |  6,500 |  40,960 |
+------+--------+---------+
|  14  | 11,500 |  81, 920|
+------+--------+---------+
|  15  | 20,000 | 163,840 |
+------+--------+---------+
|  16  | 50,000 | 327,680 |
+------+--------+---------+

"""
import argparse
from datetime import datetime
from multiprocessing import Pool
from numba import njit, uint8, uint64, int64, float64, prange
from numba.typed import List
import numpy as np

from sketchnu.hyperloglog import HyperLogLog_nu

# Empirically determined threshold provided by the paper authors
# at http://goo.gl/iU8Ig).
sub_algorithm_threshold = np.array(
    [80, 220, 400, 900, 1800, 3100, 6500, 11500, 20000, 50000], dtype=np.int64
)


def parse_cmd_line():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--n_trials",
        type=int,
        default=5000,
        help="Number of trials to run for each (p,v) pair. Default is 5000",
    )
    parser.add_argument(
        "--n_measurements",
        type=int,
        default=200,
        help="Number of measurements to do for each precision. Default is 200",
    )
    return parser.parse_args()


@njit(nogil=True)
def one_trial(keys, p, seed, lower, upper, n_measurements):
    """
    Numba function to run a single trial for the given precision level.
    Measurements for the estimated value and its bias are collected at
    `n_measurements` locations evenly spaced between [`lower`, `upper`] as the
    `keys` are added.

    Parameters
    ----------
    keys : numba.typed.List(bytes)
        Keys to be inserted into the HyperLogLog
    p : int
        Precision of the HyperLogLog
    seed : int
        Seed of the HyperLogLog
    lower : int
        Take first measurement when this many keys have been added
    upper : int
        Take last measurement when this many keys have been added
    n_measurements : int
        Number of points between [lower, upper] to record estimate & bias
    
    Returns
    -------
    raw_estimate : np.array, shape=(n_measurements,)
        Array of the raw estimate values at each of the n_measurements
    bias: np.ndarray, shape=(n_measurements,)
        Array of the measured bias values at each of the n_measurements

    """
    p = uint64(p)
    hll = HyperLogLog_nu(p, seed)
    lower = int64(lower)
    upper = int64(upper)

    measurement_points = np.linspace(lower, upper, n_measurements).astype(uint64)

    raw_estimate = np.zeros(n_measurements, float64)
    bias = np.zeros(n_measurements, float64)

    ptr = 0
    for n in range(upper):
        hll.add(keys[n])
        true_count = n + 1
        if true_count == measurement_points[ptr]:
            estimate = hll._estimation_function()
            raw_estimate[ptr] = estimate
            bias[ptr] = estimate - true_count
            ptr = ptr + 1

    return raw_estimate, bias


@njit(nogil=True, parallel=True)
def one_precision(keys, p, n_trials, lower, upper, n_measurements):
    """
    Runs `n_trials` runs for the given precision level. Measurements for the
    estimated value and its bias are collected at `n_measurements` locations
    evenly spaced between [lower, upper] as the keys are added. Note: This is
    run with `parallel=True`

    Parameters
    ----------
    keys : numba.typed.List(bytes)
        Keys to be inserted into the HyperLogLogs
    p : int
        Precision of the HyperLogLogs
    n_trials : int
        Number of independent trials to perform. Each trial will use the
        same keys, but different seeds for the HyperLogLogs
    lower : int
        Take first measurement when this many keys have been added
    upper : int
        Take last measurement when this many keys have been added
    n_measurements : int
        Number of points between [lower, upper] to record estimate & bias
    
    Returns
    -------
    raw_estimates : np.ndarray, shape=[n_trials, n_measurements]
        2-d array of the raw estimated values at each of the n_measurements
        and n_trials. Average over the n_trials to get final values
    biases : np.ndarray, shape=[n_trials, n_measurements]
        2-d arry of the bias values at each of the n_measurements and n_trials.
        Average over the n_trials to get the final values.

    """
    raw_estimates = np.zeros((n_trials, n_measurements), float64)
    biases = np.zeros((n_trials, n_measurements), float64)

    for i in prange(n_trials):
        raw_estimates[i], biases[i] = one_trial(
            keys, p, i, lower, upper, n_measurements
        )

    return raw_estimates, biases


def main():
    args = parse_cmd_line()
    n_trials = args.n_trials
    n_measurements = args.n_measurements

    # Store the mean values for each precision at each measurement
    raw_estimate = np.zeros((10, n_measurements))
    bias_data = np.zeros((10, n_measurements))

    start = datetime.now()
    for p in range(7, 17):
        start_p = datetime.now()
        print(f"{start_p}: Starting precision {p}")

        # These limits are from the HyperLogLog++ paper
        lower = sub_algorithm_threshold[p - 7]
        upper = 5 * 2 ** p

        # Make random 16-byte keys. Make sure no duplicates.
        keys = [bytes(r) for r in np.random.randint(0, 256, (upper, 16), np.uint8)]
        while len(keys) != len(set(keys)):
            keys = [bytes(r) for r in np.random.randint(0, 256, (2 ** p, 16), np.uint8)]
        keys = List(keys)

        # Run the experiments for this precision level
        # Each experiment uses a different seed value
        raw, bias = one_precision(keys, p, n_trials, lower, upper, n_measurements)
        raw_estimate[p - 7] = raw.mean(0)
        bias_data[p - 7] = bias.mean(0)
        end_p = datetime.now()
        print(f"{end_p}: Finished precision {p} in {end_p-start_p}")

    end = datetime.now()
    speed = n_trials / (end - start).total_seconds()
    print(f"{end}: Took {end-start} at {speed:.3f} trials / second")

    # Write results to a file that can easily be imported
    with open("hll_constants.py", "w") as f:
        f.write("import numpy as np\n")

        f.write("\nsub_algorithm_threshold = np.array([\n")
        f.write("    ")
        f.write(", ".join([str(i) for i in sub_algorithm_threshold]))
        f.write("\n], dtype=np.int64)\n")

        f.write("\nraw_estimate = np.array([\n")
        for i in range(7, 16):
            f.write(f"    # Precision {i}\n")
            f.write("    [")
            f.write(", ".join([f"{x:.5f}" for x in raw_estimate[i - 7, :]]))
            f.write("],\n")
        f.write(f"    # Precision 16\n")
        f.write("    [")
        f.write(", ".join([f"{x:.5f}" for x in raw_estimate[-1, :]]))
        f.write("]\n")
        f.write("])\n")

        f.write("\nbias_data = np.array([\n")
        for i in range(7, 16):
            f.write(f"    # Precision {i}\n")
            f.write("    [")
            f.write(", ".join([f"{x:.5f}" for x in bias_data[i - 7, :]]))
            f.write("],\n")
        f.write(f"    # Precision 16\n")
        f.write("    [")
        f.write(", ".join([f"{x:.5f}" for x in bias_data[-1, :]]))
        f.write("]\n")
        f.write("])\n")


if __name__ == "__main__":
    main()



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
import numpy as np
from pathlib import Path
import pytest

from sketchnu.hyperloglog import HyperLogLog

n_keys = 50000
n_trials = 20
keys = list(
    set([bytes(k) for k in np.random.randint(0, 256, (n_keys, 8)).astype(np.uint8)])
)
n_keys = len(keys)


def test_empty(p: int = 10):
    """
    Test that empty HyperLogLog returns 0.0

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is 10
    """
    hll = HyperLogLog(p)
    assert hll.query() == 0.0


def test_repeat_keys(p: int = 10):
    """
    Test that estimated count does change when you put the same set of keys in
    """
    hll = HyperLogLog(p)
    hll.update(keys)
    n_uniq = hll.query()
    hll.update(keys)
    assert n_uniq == hll.query()


def test_update_low(p: int = 10, n_trials: int = 20, t_stat: float = 3.291):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    We insert a random number of keys between [5, threshold - 5]. We repeat this
    test n_trials times using different seeds. The query() should equal the number
    of unique keys that were added to the HyperLogLog.

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is a relative low value of 10 in order
        to lower the threshold value for faster testing. Default is 10
    n_trials : int, optional
        Number of different experiments to run. Each experiment uses a different
        seed to initialize a HyperLogLog. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    hll = HyperLogLog(p)
    error = np.zeros(n_trials)
    n_uniq = np.random.randint(5, hll.threshold - 5)
    n_uniq = min(n_uniq, n_keys)
    for seed in range(n_trials):
        hll = HyperLogLog(p, seed)
        hll.update(keys[:n_uniq])
        error[seed] = n_uniq - hll.query()

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean != 0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
        # 99.9% confidence interval
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_update_med(p: int = 10, n_trials: int = 20, t_stat: float = 3.291):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    We insert a random number of keys between [threshold+20, 5 * 2^p]. We repeat
    this test n_trials times using different seeds. The query() should equal the
    number of unique keys that were added to the HyperLogLog.

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is a relative low value of 10 in order
        to lower the threshold value for faster testing. Default is 10
    n_trials : int, optional
        Number of different experiments to run. Each experiment uses a different
        seed to initialize a HyperLogLog. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    hll = HyperLogLog(p)
    error = np.zeros(n_trials)
    n_uniq = np.random.randint(hll.threshold + 20, 5 * (2 ** p))
    n_uniq = min(n_uniq, n_keys)
    for seed in range(n_trials):
        hll = HyperLogLog(p, seed)
        hll.update(keys[:n_uniq])
        error[seed] = n_uniq - hll.query()

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean != 0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
        # 99.9% confidence interval
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_update_hi(p: int = 10, n_trials: int = 20, t_stat: float = 3.291):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    We insert all the keys which tests the algorithm above the 5 * 2^p limit.
    We repeat this test n_trials times using different seeds. The query() should
    equal the number of unique keys that were added to the HyperLogLog.

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is a relative low value of 10 in order
        to lower the threshold value for faster testing. Default is 10
    n_trials : int, optional
        Number of different experiments to run. Each experiment uses a different
        seed to initialize a HyperLogLog. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    error = np.zeros(n_trials)
    assert n_keys > (5 * (2 ** p))
    n_uniq = n_keys
    for seed in range(n_trials):
        hll = HyperLogLog(p, seed)
        hll.update(keys[:n_uniq])
        error[seed] = n_uniq - hll.query()

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean != 0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
        # 99.9% confidence interval
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_update_ngram(p: int = 16):
    """
    Test that ngraming is done right

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is 16
    """
    data = b"0123456789abcdef"
    for ngram in range(1, len(data) + 1):
        hll = HyperLogLog(p)
        hll.update_ngram([data], ngram)
        true_count = len(data) - ngram + 1
        assert hll.query() == pytest.approx(true_count, rel=1e-3)


def test_merge(p: int = 10, n_trials: int = 20, t_stat: float = 3.291):
    """
    Uses a t-test to test the null hypothesis that the mean of the
    difference between the true count and estimated count is 0. Uses a
    confidence level of 99.9% to reject the null hypotheses. The test asserts
    that we should fail to reject the null hypothesis.

    We insert `n_keys` random keys into the first HyperLogLog and `n_keys`
    different random keys into a second HyperLogLog. We merge the second into the
    first and then check that the estimated count is equal to the size of the union
    of the two sets of random keys.

    We repeat this test n_trials times using different seeds. The query() should
    equal the number of unique keys that were added to the HyperLogLog.

    Parameters
    ----------
    p : int, optional
        Precision of the HyperLogLog. Default is a relative low value of 10 in order
        to lower the threshold value for faster testing. Default is 10
    n_trials : int, optional
        Number of different experiments to run. Each experiment uses a different
        seed to initialize a HyperLogLog. Default is 20
    t_stat : float, optional
        t-test statistic which sets the confidence interval. Default corresponds
        to 99.9%. Default is 3.291
    """
    keys2 = list(
        set([bytes(k) for k in np.random.randint(0, 256, (n_keys, 8)).astype(np.uint8)])
    )
    n_uniq = len(set(keys2).union(set(keys)))

    error = np.zeros(n_trials)
    for seed in range(n_trials):
        hll1 = HyperLogLog(p, seed)
        hll2 = HyperLogLog(p, seed)
        hll1.update(keys)
        hll2.update(keys2)
        hll1.merge(hll2)
        error[seed] = n_uniq - hll1.query()

    if error.std() == 0.0:
        assert error.mean() == 0.0, f"std = 0, but mean !=0, {error.mean():.3f}"
    else:
        t_value = np.abs(error.mean() / (error.std() / np.sqrt(n_trials)))
        # 99.9% confidence level
        assert t_value < t_stat, f"t-value {t_value:.4} is above {t_stat}"


def test_saving(tmp_path: Path, p: int = 10):
    """
    Test saving and loading a HyperLogLog

    Parameters
    ----------
    tmp_path : Path
        This is a pytest fixture which will provide a temporary directory unique
        to the test invocation, created in the base temporary directory
    p : int, optional
        Precision of the HyperLogLog. Default is 10
    """
    hll = HyperLogLog(p)
    hll.update(keys)
    filename = tmp_path / "test_hll.npz"

    hll.save(filename)
    hll_load = HyperLogLog.load(filename)

    assert hll.query() == hll_load.query()


if __name__ == "__main__":
    test_empty()
    test_repeat_keys()
    test_update_low()
    test_update_med()
    test_update_hi()
    test_merge()



"""
Sketchnu has Numba implementations of sketch algorithms and other useful functions
that utilize hash functions.

Copyright (C) 2022 Matthew Hendrey

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
from collections import Counter
import numpy as np
from pathlib import Path

from sketchnu.heavyhitters import HeavyHitters
from sketchnu.countmin import CountMin
from .zipf import zipf


vocab_size = 50000
vocab = list(
    set([bytes(k) for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)])
)
zipf_p = zipf(vocab_size)
n_data = 50000
data_stream = np.random.choice(vocab, n_data, p=zipf_p).tolist()


def test_n_added(width: int = 100, depth: int = 4, max_key_len: int = 8):
    """
    Test that number of elements added to the sketch is properly tracked

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 100
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    hh = HeavyHitters(width, depth, max_key_len)
    hh.update(data_stream)
    assert hh.n_added() == n_data


def test_update_theory(width: int = 500, depth: int = 4, max_key_len: int = 8):
    """
    Tests both lemmas in the original paper (3.4.3 & 3.4.4). This uses the theoretical
    work which includes using a count-min sketch to estimate frequency of occurrence.
    In practice, following advice from the paper, you can drop the count-min sketch and
    use the lhh_count as an estimate. The lhh_count <= true_count <= cms estimate

    Lemma 3.4.3
    -----------
    Topkapi with depth = log(2/delta) & width = 1/eps where eps < phi misses to report
    key with frequency >= phi * N with probability at most delta / 2

    Lemma 3.4.4
    -----------
    Topkapi with depth = log(2/delta) & width = 1/eps where eps < phi reports key with
    frequency <= (phi - eps) * N with probability at most delta / 2

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    eps = 1 / width
    phi = 1.1 * eps
    threshold = int(n_data * phi)
    failure_rate = np.exp(-depth)
    failure_miss = []
    failure_reports = []

    hh = HeavyHitters(width, depth, max_key_len, phi)
    cms = CountMin("linear", width, depth)

    hh.update(data_stream)
    cms.update(data_stream)
    true_count = Counter(data_stream)

    # This is the full Topkapi algorithm. See Sec. 3.2 of the paper
    # Consider the union of all keys in LHH as candidate set. Estimate their counts
    # using the CMS & report all elements with count higher than phi * N
    hh.generate_candidate_set(0)  # Puts union of all keys into candidate_set
    full_approx = Counter()
    for k, v in hh.candidate_set.items():
        if cms[k] >= threshold:
            full_approx[k] = cms[k]

    for v in true_count:
        # Misses to report when it should
        if true_count[v] >= (phi * n_data):
            if v in full_approx:
                failure_miss.append(0)
            else:
                failure_miss.append(1)
        # Reports when it shouldn't
        if true_count[v] <= (phi - eps) * n_data:
            if v in full_approx:
                failure_reports.append(1)
            else:
                failure_reports.append(0)

    if len(failure_miss) > 0:
        assert np.mean(failure_miss) <= failure_rate
    if len(failure_reports) > 0:
        assert np.mean(failure_reports) <= failure_rate


def test_update_ideal(width: int = 2500, depth: int = 4, max_key_len: int = 8):
    """
    Setting width >> vocab_size to make sure the machinery is working as expected.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 2500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    vocab_size = 25
    vocab = list(
        set(
            [
                bytes(k)
                for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)
            ]
        )
    )
    zipf_p = zipf(vocab_size)
    n_data = 50000
    data_stream = np.random.choice(vocab, n_data, p=zipf_p).tolist()

    hh = HeavyHitters(width, depth, max_key_len)
    hh.update(data_stream)
    true_count = Counter(data_stream)

    for key, value in true_count.items():
        assert value == hh[key]


def test_update(width: int = 500, depth: int = 4, max_key_len: int = 8):
    """
    Test that keys returned by HeavyHitters.query() all appear within the set of
    keys where the true count of each is >= n_data / width.

    Because HeavyHitters uses the lhh_count instead of count-min sketch, following the
    practical advice in the paper, when HeavyHitters keeps only keys with lhh_count >=
    threshold the true count of those keys is greater than or equal to the sketch's
    estimate. So this means that we aren't getting all the keys we should and thus all
    of the sketch's keys should be found in the set of keys whose true counts are
    greater than or equal to the threshold.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    eps = 1 / width
    threshold = int(n_data * eps)

    hh = HeavyHitters(width, depth, max_key_len)

    hh.update(data_stream)
    true_count = Counter(data_stream)

    hh_set = set([k for k, _ in hh.query(vocab_size)])
    true_set = set([k for k, v in true_count.items() if v >= threshold])

    assert hh_set.intersection(true_set) == hh_set


def test_merge_ideal(width: int = 2500, depth: int = 4, max_key_len: int = 8):
    """
    Add same data stream into two sketches. Merge them.

    Setting width >> vocab_size to make sure the machinery is working as expected.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 2500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    vocab_size = 25
    vocab = list(
        set(
            [
                bytes(k)
                for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)
            ]
        )
    )
    zipf_p = zipf(vocab_size)
    n_data = 50000
    data_stream = np.random.choice(vocab, n_data, p=zipf_p).tolist()

    hh1 = HeavyHitters(width, depth, max_key_len)
    hh2 = HeavyHitters(width, depth, max_key_len)

    hh1.update(data_stream)
    hh2.update(data_stream)
    true_count = Counter(data_stream)
    true_count.update(data_stream)

    hh1.merge(hh2)
    for key, value in true_count.items():
        assert value == hh1[key]


def test_merge_mixed_key_len(width: int = 2500, depth: int = 4, max_key_len: int = 8):
    """
    Test that we can handle the merging of different length keys.

    Setting width >> vocab_size to make sure the machinery is working as expected.

    Parameters
    ----------
    width : int, optional
        Width of the sketch. Default is 2500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    """
    assert max_key_len - 3 > 0, "max_key_len is too small must be > 3"

    vocab_size = 25
    vocab = list(
        set(
            [
                bytes(k)
                for k in np.random.randint(0, 256, (vocab_size, 8)).astype(np.uint8)
            ]
        )
    )
    zipf_p = zipf(vocab_size)
    n_data = 50000
    data_stream = np.random.choice(vocab, n_data, p=zipf_p).tolist()

    vocab2 = list(
        set(
            [
                bytes(k)
                for k in np.random.randint(
                    0, 256, (vocab_size, max_key_len - 3)
                ).astype(np.uint8)
            ]
        )
    )
    zipf_p2 = zipf(vocab_size)
    data_stream2 = np.random.choice(vocab2, n_data, p=zipf_p2).tolist()

    hh1 = HeavyHitters(width, depth, max_key_len)
    hh2 = HeavyHitters(width, depth, max_key_len)

    hh1.update(data_stream)
    hh2.update(data_stream2)
    true_count = Counter(data_stream)
    true_count.update(data_stream2)

    hh1.merge(hh2)
    for key, value in true_count.items():
        assert value == hh1[key]

    # Reverse the ordering
    hh1 = HeavyHitters(width, depth, max_key_len)
    hh2 = HeavyHitters(width, depth, max_key_len)

    hh1.update(data_stream2)
    hh2.update(data_stream)

    hh1.merge(hh2)
    for key, value in true_count.items():
        assert value == hh1[key]


def test_saving(
    tmp_path: Path,
    width: int = 500,
    depth: int = 4,
    max_key_len: int = 8,
    topk: int = 10,
):
    """
    Test saving and loading the sketch

    Parameters
    ----------
    tmp_path : Path
        This is a pytest fixture which will provide a temporary directory unique
        to the test invocation, created in the base temporary directory
    width : int, optional
        Width of the sketch. Default is 500
    depth : int, optional
        Depth of the sketch. Default is 4
    max_key_len : int, optional
        Maximum length of any key inserted into the sketch. Default is 8
    topk : int, optional
        Query for the topk most frequent. Default is 10
    """
    hh = HeavyHitters(width, depth, max_key_len)
    hh.update(data_stream)
    filename = tmp_path / "test_hh.npz"

    hh.save(filename)
    hh_load = HeavyHitters.load(filename)

    for orig, saved in zip(hh.query(topk), hh_load.query(topk)):
        assert orig == saved
